# Copyright (c) Meta Platforms, Inc. and affiliates.

# All rights reserved.

# This source code is licensed under the license found in the
# LICENSE file in the root directory of this source tree.

from functools import partial

import torch
import torch.nn as nn
import torch.nn.functional as F
from timm.models.layers import trunc_normal_, DropPath
from torch import Tensor
from collections import OrderedDict
from torchvision.ops.feature_pyramid_network import FeaturePyramidNetwork
from typing import Optional, Union, List, Dict


class Block(nn.Module):
    r""" ConvNeXt Block. There are two equivalent implementations:
    (1) DwConv -> LayerNorm (channels_first) -> 1x1 Conv -> GELU -> 1x1 Conv; all in (N, C, H, W)
    (2) DwConv -> Permute to (N, H, W, C); LayerNorm (channels_last) -> Linear -> GELU -> Linear; Permute back
    We use (2) as we find it slightly faster in PyTorch

    Args:
        dim (int): Number of input channels.
        drop_path (float): Stochastic depth rate. Default: 0.0
        layer_scale_init_value (float): Init value for Layer Scale. Default: 1e-6.
    """

    def __init__(self, dim: int, drop_path: Optional[float] = 0.0,
                 layer_scale_init_value: Optional[float] = 1e-6) -> None:
        super().__init__()
        self.dwconv = nn.Conv2d(dim, dim, kernel_size=(7, 7), padding=3, groups=dim)  # depthwise conv
        self.norm = LayerNorm(dim, eps=1e-6)
        self.pwconv1 = nn.Linear(dim, 4 * dim)  # pointwise/1x1 convs, implemented with linear layers
        self.act = nn.GELU()
        self.pwconv2 = nn.Linear(4 * dim, dim)
        self.gamma = nn.Parameter(layer_scale_init_value * torch.ones((dim)),
                                  requires_grad=True) if layer_scale_init_value > 0 else None
        self.drop_path: Union[DropPath, nn.Identity] = DropPath(drop_path) if drop_path > 0. else nn.Identity()

    def forward(self, x: Tensor) -> Tensor:  # Image of shape (B, C, H, W) -> (10, 3, 100, 100)
        input = x
        x = self.dwconv(x)
        x = x.permute(0, 2, 3, 1)  # (N, C, H, W) -> (N, H, W, C)
        x = self.norm(x)
        x = self.pwconv1(x)
        x = self.act(x)
        x = self.pwconv2(x)
        if self.gamma is not None:
            x = self.gamma * x
        x = x.permute(0, 3, 1, 2)  # (N, H, W, C) -> (N, C, H, W)

        x = input + self.drop_path(x)
        return x


class ConvNeXt(nn.Module):
    r""" ConvNeXt
        A PyTorch impl of : `A ConvNet for the 2020s`  -
          https://arxiv.org/pdf/2201.03545.pdf

    Args:
        in_channels (int): Number of input image channels. Default: 3
        depths (tuple(int)): Number of blocks at each stage. Default: [3, 3, 9, 3]
        dims (int): Feature dimension at each stage. Default: [96, 192, 384, 768]
        drop_path_rate (float): Stochastic depth rate. Default: 0.
        layer_scale_init_value (float): Init value for Layer Scale. Default: 1e-6.
    """

    def __init__(self, in_channels: Optional[int] = 3,
                 depths: Optional[List[int]] = [3, 3, 9, 3],
                 dims: Optional[List[int]] = [96, 192, 384, 768],
                 drop_path_rate: Optional[float] = 0.0, layer_scale_init_value: Optional[float] = 1e-6,
                 out_indices: Optional[List[int]] = [0, 1, 2, 3],
                 out_channels: int = 256
                 ):
        super().__init__()

        self.downsample_layers = nn.ModuleList()  # stem and 3 intermediate downsampling conv layers
        self.stages = nn.ModuleList()  # 4 feature resolution stages, each consisting of multiple residual blocks
        self.norms = nn.ModuleList()
        self.out_indices = out_indices

        dp_rates = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]
        cur = 0

        self.fpn = FeaturePyramidNetwork(
            in_channels_list=dims,
            out_channels=out_channels,
            extra_blocks=None,
        )

        # Create the network Stem
        stem = nn.Sequential(
            nn.Conv2d(in_channels, dims[0], kernel_size=(4, 4), stride=(4, 4)),
            LayerNorm(dims[0], eps=1e-6, data_format="channels_first")
        )
        self.downsample_layers.append(stem)

        # Downsample layers
        for i in range(3):
            downsample_layer = nn.Sequential(
                LayerNorm(dims[i], eps=1e-6, data_format="channels_first"),
                nn.Conv2d(dims[i], dims[i + 1], kernel_size=(2, 2), stride=(2, 2)),
            )
            self.downsample_layers.append(downsample_layer)

        # Stages
        for i in range(4):
            stage = nn.Sequential(
                *[Block(dim=dims[i], drop_path=dp_rates[cur + j],
                        layer_scale_init_value=layer_scale_init_value) for j in range(depths[i])]
            )
            self.stages.append(stage)
            cur += depths[i]

        # Layer Normalization
        norm_layer = partial(LayerNorm, eps=1e-6, data_format="channels_first")
        for i_layer in range(4):
            layer = norm_layer(dims[i_layer])
            self.norms.append(layer)

        self.apply(self._init_weights)

    def _init_weights(self, m):
        if isinstance(m, (nn.Conv2d, nn.Linear)):
            trunc_normal_(m.weight, std=.02)
            nn.init.constant_(m.bias, 0)

    def init_weights(self, pretrained=None):
        """Initialize the weights in backbone.
        Args:
            pretrained (str, optional): Path to pre-trained weights.
                Defaults to None.
        """

        def _init_weights(m):
            if isinstance(m, nn.Linear):
                trunc_normal_(m.weight, std=.02)
                if isinstance(m, nn.Linear) and m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.bias, 0)
                nn.init.constant_(m.weight, 1.0)

    def forward_features(self, x: Tensor) -> Dict[str, Tensor]:
        out = OrderedDict()
        for i, (down, stage, norm) in enumerate(zip(self.downsample_layers, self.stages, self.norms)):
            x = down(x)
            x = stage(x)
            x = norm(x)
            out[str(i)] = x
        return out

    def forward(self, x):
        x = self.forward_features(x)
        x = self.fpn(x)
        return x


class LayerNorm(nn.Module):
    r""" LayerNorm that supports two data formats: channels_last (default) or channels_first.
    The ordering of the dimensions in the inputs. channels_last corresponds to inputs with
    shape (batch_size, height, width, channels) while channels_first corresponds to inputs
    with shape (batch_size, channels, height, width).
    """

    def __init__(self, normalized_shape, eps=1e-6, data_format="channels_last"):
        super().__init__()
        self.weight = nn.Parameter(torch.ones(normalized_shape))
        self.bias = nn.Parameter(torch.zeros(normalized_shape))
        self.eps = eps
        self.data_format = data_format
        if self.data_format not in ["channels_last", "channels_first"]:
            raise NotImplementedError
        self.normalized_shape = (normalized_shape,)

    def forward(self, x: Tensor) -> Tensor:
        if self.data_format == "channels_last":
            x = F.layer_norm(x, self.normalized_shape, self.weight, self.bias, self.eps)
        elif self.data_format == "channels_first":
            u = x.mean(1, keepdim=True)
            s = (x - u).pow(2).mean(1, keepdim=True)
            x = (x - u) / torch.sqrt(s + self.eps)
            x = self.weight[:, None, None] * x + self.bias[:, None, None]
        return x



import torch.nn as nn
import torch
from torch import Tensor
from typing import List, Tuple, Optional
from hcat.models.unext_modules import LayerNorm, Block, DropPath, r_block, UpSampleLayer, ConcatConv, trunc_normal_
from functools import partial


@torch.jit.script
def _crop(img: Tensor, x: int, y: int, z: int, w: int, h: int, d: int) -> Tensor:
    return img[..., x:x + w, y:y + h, z:z + d]


class UNeXT(nn.Module):
    def __init__(self, in_channels: Optional[int] = 1,
                 depths: Optional[List[int]] = (1, 2, 3, 2, 1),  # [1, 2, 3, 2, 1],
                 dims: Optional[List[int]] = (32, 64, 128, 64, 32),  # [16, 32, 64, 32, 16],
                 drop_path_rate: Optional[float] = 0.0,
                 layer_scale_init_value: Optional[float] = 1.,
                 out_channels: int = 4,
                 kernel_size: int = 7,
                 activation: Optional = nn.GELU,
                 normalization: Optional = LayerNorm
                 ):
        super(UNeXT, self).__init__()

        assert len(depths) == len(dims), f'Number of depths should equal number of dims: {depths} != {dims}'

        # ----------------- Save Init Params for Preheating...
        self.dims = dims
        self.depths = depths
        self.in_channels = in_channels
        self.out_channels = out_channels

        # ----------------- Model INIT
        self.downsample_layers = nn.ModuleList()  # stem and 3 intermediate downsampling conv layers
        self.upsample_layers = nn.ModuleList()  # stem and 3 intermediate downsampling conv layers

        # Drop path should not decrease to nothing over time...
        dp_rates = [drop_path_rate for _ in range(sum(depths))]

        self.down_stages = nn.ModuleList()  # 4 feature resolution stages, each consisting of multiple residual blocks
        self.up_stages = nn.ModuleList()  # 4 feature resolution stages, each consisting of multiple residual blocks

        self.down_norms = nn.ModuleList()
        self.up_norms = nn.ModuleList()

        # ----------------- Stage 1
        self.init_stage = nn.Sequential(
            nn.Conv3d(in_channels, dims[0], kernel_size=7, padding=3),
            Block(dim=dims[0], drop_path=0.0, kernel_size=kernel_size, activation=activation)
        )

        # ----------------- Unlike original conv next we cannot reduce the stem by that much...
        stem = nn.Sequential(
            nn.Conv3d(dims[0], dims[0], kernel_size=(2, 2, 2), stride=(2, 2, 2)),
            LayerNorm(dims[0], eps=1e-6, data_format="channels_first")
        )
        self.downsample_layers.append(stem)

        # ----------------- Downsample layers
        for i in range(len(dims) // 2):
            downsample_layer = nn.Sequential(
                LayerNorm(dims[i], eps=1e-6, data_format="channels_first"),
                nn.Conv3d(dims[i], dims[i + 1], kernel_size=(2, 2, 2), stride=(2, 2, 2)),
            )
            self.downsample_layers.append(downsample_layer)

        # Down Stages
        cur = 0
        for i in range(len(dims) // 2 + 1):
            num_blocks = depths[i]
            stage = nn.Sequential(
                *[Block(dim=dims[i], drop_path=dp_rates[cur + j], kernel_size=kernel_size, activation=activation,
                        layer_scale_init_value=layer_scale_init_value) for j in range(num_blocks)]
            )
            self.down_stages.append(stage)
            cur += depths[i]

        # Down Layer Normalization
        norm_layer = partial(LayerNorm, eps=1e-6, data_format="channels_first")
        for i_layer in range(len(dims) // 2 + 1):
            layer = norm_layer(dims[i_layer])
            self.down_norms.append(layer)

        # Up Layer Normalization
        norm_layer = partial(LayerNorm, eps=1e-6, data_format="channels_first")
        for i_layer in range(len(dims) // 2):
            layer = norm_layer(dims[i_layer + len(dims) // 2 + 1])
            self.up_norms.append(layer)

        # Upsample layers
        for i in range(len(dims) // 2):
            upsample_layer = UpSampleLayer(dims[i + len(dims) // 2], dims[i + 1 + len(dims) // 2])
            self.upsample_layers.append(upsample_layer)

        # Up Stages
        cur = 0
        for i in range(len(dims) // 2):
            num_blocks = depths[i + len(dims) // 2 + 1]
            stage = nn.Sequential(
                *[Block(dim=dims[i + len(dims) // 2 + 1], drop_path=dp_rates[cur + j], kernel_size=kernel_size,
                        activation=activation,
                        layer_scale_init_value=layer_scale_init_value) for j in range(num_blocks)]
            )
            self.up_stages.append(stage)
            cur += depths[i]

        # Concat Layers
        self.concat = nn.ModuleList()
        for i in range(len(dims) // 2):
            self.concat.append(ConcatConv(dims[i + len(dims) // 2 + 1]))

        self.final_concat = ConcatConv(in_channels=dims[-1])

        # Final Upscale
        self.upscale_out = UpSampleLayer(dims[-1], dims[-1])

        self.sigmoid = nn.Sigmoid()
        self.tanh = nn.Tanh()

        # Flatten Channels to out...
        # self.out_conv = nn.Conv3d(dims[-1], out_channels, kernel_size=1)
        self.out_embed = nn.Conv3d(dims[-1], 3, kernel_size=3, stride=1, dilation=1, padding=1)
        self.out_prob = nn.Conv3d(dims[-1], 1, kernel_size=3, stride=1, dilation=1, padding=1)

        # self.apply(self._init_weights)

    def forward(self, x: Tensor) -> Tensor:
        """

        NEEDS INPUT TO BE NORMALIZED TO GET ANYWHERE! WAS JACK SHIT BEFORE NORMALIZATION!!!!

        Plan is to preinitalize steps to include the necessary shit with torch.empty(devie=cuda)
        I think that allocating and dealocating is expensive
        See significant speedup in epoch time if batch size is 1

        I dont know why... Figure out tomorrow. Could be data bus and distributed data?

        Time per epoch with batch size of 24: ~30 seconds
        Time per epoch with batch size of 2: ~10-15seconds

        :param x:
        :return:
        """
        x = self.init_stage(x)

        shapes: List[List[int]] = [x.shape]
        steps: List[Tensor] = [x]

        assert len(self.upsample_layers) == len(self.concat)
        assert len(self.upsample_layers) == len(self.up_stages)
        assert len(self.upsample_layers) == len(self.up_norms)

        for i, (down, stage, norm) in enumerate(zip(self.downsample_layers, self.down_stages, self.down_norms)):

            # Save input for upswing of Unet
            if i != 0:
                steps.append(x)

            x = down(x)
            x = norm(x)
            x = stage(x)
            shapes.append(x.shape)

        shapes.reverse()
        for i, (up, cat, stage, norm) in enumerate(
                zip(self.upsample_layers, self.concat, self.up_stages, self.up_norms)):
            x = up(x, shapes[i + 1])
            y = steps.pop(-1)
            x = cat(x, y)
            x = stage(x)
            x = norm(x)

        # Out

        x = self.upscale_out(x, shapes[-1])
        y = steps.pop(-1)
        x = self.final_concat(x, y)
        # x = self.out_conv(x)

        x = torch.cat((
            self.tanh(self.out_embed(x)),
            self.sigmoid(self.out_prob(x))
        ), dim=1)

        return x

    def _init_weights(self, m):
        if isinstance(m, nn.Linear):
            trunc_normal_(m.weight, std=.5)
            if isinstance(m, nn.Linear) and m.bias is not None:
                nn.init.constant_(m.bias, 0)
        elif isinstance(m, nn.LayerNorm):
            nn.init.constant_(m.bias, 0.0)
            nn.init.constant_(m.weight, 0.01)
        # elif isinstance(m, nn.Conv3d):
        #     trunc_normal_(m.weight, std=0.5)
        #     nn.init.constant_(m.bias, 0.0)

    def init_weights(self, pretrained=None):
        """Initialize the weights in backbone.
        Args:
            pretrained (str, optional): Path to pre-trained weights.
                Defaults to None.
        """

        def _init_weights(m):
            if isinstance(m, nn.Linear):
                trunc_normal_(m.weight, std=.02)
                if isinstance(m, nn.Linear) and m.bias is not None:
                    nn.init.constant_(m.bias, 0)
            elif isinstance(m, nn.LayerNorm):
                nn.init.constant_(m.bias, 0)
                nn.init.constant_(m.weight, 1.0)

        for m in self.parameters():
            _init_weights(m)

    def __repr__(self):
        return f'nn.Module[name=UNeXt, dims={self.dims}, depths={self.depths}, ' \
               f'in_channels={self.in_channels}, out_channels={self.out_channels}]'


@torch.jit.script
def crop(x: Tensor, shape: List[int]) -> Tensor:
    return _crop(x, 0, 0, 0, int(shape[2]), int(shape[3]), int(shape[4]))



import gc
import os.path
from itertools import product
from typing import Optional, List

import numpy as np
import psutil
import torch
from torch import Tensor
from tqdm import tqdm

import hcat
import hcat.lib.functional as functional
import hcat.lib.utils as utils
from hcat.backends.spatial_embedding import SpatialEmbedding
from hcat.backends.unet_and_watershed import UNetWatershed
from hcat.lib.cell import Cell
from hcat.lib.cochlea import Cochlea
from hcat.lib.explore_lif import get_xml
from hcat.lib.utils import calculate_indexes, remove_edge_cells, remove_wrong_sized_cells


# DOCUMENTED

def _segment(f: str, channel: int, intensity_reject_threshold: float,
             dtype: str = 'uint16',
             unet: bool = False, cellpose: bool = False, no_post: bool = False,
             figure: Optional[bool] = True) -> Cochlea:
    """

    Segment arbitrarily large images of cochleae.
    Loads a lif or tif file and iterativley segments all hair cells in the image.
    Saves the results of the analysis in a file with extension ".cochlea" which compresses and saves the analysis.

    Example:

    >>> from hcat.segment import _segment
    >>> f = 'path/to/my/file.tif'
    >>> segment(f, intensity_reject_threshold=0.5)

    :param *str* f: path to file
    :param *int* channel: index of cytosolic stain (0 if only one channel).
    :param *float* intensity_reject_threshold: predicted cell masks with mean cytosolic stain below this threshold will be rejected
    :param *str,optional* dtype: bit depth of input image. Possible values: uint8, uint16, int8, int16.
    :param *bool,optional* unet: Perform segmentation with UNet + Watershed segmentation backbone.
    :param *bool,optinal* cellpose: Perform segmentation with 3D cellpose segmentation backbone.
    :param *bool,optinal* no_post: if true, remove cell postprocessing, which may remove legitimate cells.
    :param *bool,optinal* figure: If true, renders a diagnostic figure for this cochlea.
    :return: None
    """
    print('Initializing hair cell segmentation algorithm...')
    if f is None:
        print('\x1b[1;31;40m' + 'ERROR: No File to Analyze... \nAborting.' + '\x1b[0m')
        return None

    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    if device == 'cuda':
        print('\x1b[1;32;40m' + 'CUDA: GPU successfully initialized!' + '\x1b[0m')
    else:
        print(
            '\x1b[1;33;40m' + 'WARNING: GPU not present or CUDA is not correctly intialized for GPU accelerated computation. '
                              'Analysis may be slow.' + '\x1b[0m')

    max_cell: int = 0
    cells = []
    append_cells = cells.append
    intensity_reject: callable = torch.jit.script(
        hcat.lib.functional.IntensityCellReject(threshold=intensity_reject_threshold).requires_grad_(False))
    predict_curvature: callable = hcat.lib.functional.PredictCurvature(erode=3)

    with torch.no_grad():

        backend = _get_backend(unet, cellpose, device)  # may return none if failed
        if backend is None:
            return None

        image_base = utils.load(f, 'TileScan 1 Merged', verbose=True)  # may return none if failed
        print('SEG SHAPE', image_base.shape)
        if image_base is None:
            return None

        print(
            f'DONE: shape: {image_base.shape}, min: {image_base.min()}, max: {image_base.max()}, dtype: {image_base.dtype}')

        dtype = image_base.dtype if dtype is None else dtype
        scale: int = hcat.lib.utils.get_dtype_offset(dtype)

        c, x, y, z = image_base.shape
        out_img = torch.zeros((1, x, y, z), dtype=torch.long)  # (1, X, Y, Z)

        x_ind: List[List[int]] = calculate_indexes(15, 512, x, x)  # [[0, 255], [30, 285], ...]
        y_ind: List[List[int]] = calculate_indexes(15, 512, y, y)  # [[0, 255], [30, 285], ...]
        total: int = len(x_ind) * len(y_ind)

        for (x, y) in tqdm(product(x_ind, y_ind), total=total, desc='Segmenting: '):

            # Load and prepare image for ML model evaluation
            image: np.array = image_base[:, x[0]:x[1], y[0]:y[1]].astype(np.float)
            image: Tensor = torch.from_numpy(image).float().div(scale).unsqueeze(0)
            # image: Tensor = image.div(0.5) if image.min() < 0 else image.sub(0.5).div(0.5)  # signed or unsigned?

            # which backend do we use?
            out: Tensor = backend(image.to(device)) if unet else backend(image[:, [channel], ...].to(device))

            # backend will return an empty tensor if there wasn't usable data underneath
            if out.shape[1] == 0:
                continue

            # postprocess
            out: Tensor = intensity_reject(out, image[:, [channel], ...].to(device)) if not no_post else out

            # backend will return an empty tensor if there wasn't usable data underneath
            if out.shape[1] == 0:
                continue

            value, out = out.max(1)
            out[value < 0.25] = 0

            # postprocess
            # out: Tensor = remove_wrong_sized_cells(out) if not no_post else out

            # Remove cells on the edge for proper merging
            out: Tensor = remove_edge_cells(out).cpu()

            if out.shape[1] == 0:
                continue

            out[out > 0] = out[out > 0] + max_cell
            max_cell = out.max() if out.max() > 0 else max_cell

            # merging!
            x_exp = torch.tensor([x[0], x[1]])
            y_exp = torch.tensor([y[0], y[1]])
            x_exp[x_exp < 0] = 0
            y_exp[y_exp < 0] = 0
            destination = out_img[:, x_exp[0]:x_exp[1], y_exp[0]:y_exp[1], :]
            destination, new_ids = functional.merge_regions(destination.to(out.device), out, 0.35)

            out_img[:, x_exp[0]:x_exp[1], y_exp[0]:y_exp[1], :] = destination.cpu()

            for u in new_ids:
                if u == 0:
                    continue
                center = (out == u).nonzero(as_tuple=False).float().mean(0)
                center[1] += x[0]
                center[2] += y[0]
                append_cells(Cell(id=u,
                                  image=torch.clone(image[:, :, :-1:, :-1:, 0:out.shape[-1]].mul(0.5).add(0.5)),
                                  mask=(out == u).unsqueeze(0),
                                  loc=center))
        print('Predicted Number of Cells: ', len(cells))

        mem = psutil.virtual_memory()  # get memory for later
        del image_base  # memory issues are tight. removing image and image_base help
        gc.collect()

        # Not every input will be a whole cochlea... curvature means nothing over such a small region
        curvature, distance, apex = predict_curvature(out_img.sum(-1).unsqueeze(-1), cells, None)

        if curvature is None:
            print('\x1b[1;33;40mWARNING: ' +
                  'All three methods to predict hair cell path have failed. Frequency Mapping functionality is limited.'
                  ' Consider Manual Calculation.'
                  + '\x1b[0m')
        # curvature estimation really only works if there is a lot of tissue...
        if distance is not None and distance.max() > 4000:
            for c in cells: c.calculate_frequency(curvature[[0, 1], :], distance)  # calculate cell's best frequency
            # cells = [c for c in cells if not c._distance_is_far_away]  # remove a cell if its far away from curve
        else:
            curvature, distance, apex = None, None, None
            print('\x1b[1;33;40mWARNING: ' +
                  'Predicted Cochlear Distance is below 4000um. Not sufficient information to determine cell frequency.'
                  + '\x1b[0m')

        xml = get_xml(f) if f.endswith('.lif') else None

        # Store
        c = Cochlea(mask=out_img,
                    filename=f,
                    leica_metadata=xml,
                    analysis_type='segment',
                    im_shape=out_img.shape,
                    cochlear_distance=distance,
                    curvature=curvature,
                    cells=cells,
                    apex=apex)

        # This takes too much memory for huge images
        # Check the size of the outimg. If we have twice the available free memory we can compress!
        if mem.available > (out_img.numpy().size * out_img.numpy().itemsize * 2):
            c.compress_mask(vebose=True)
        else:
            print('\x1b[1;33;40m' + 'WARNING: Low free system memory. Cannot compress output... ' + '\x1b[0m')

        c.save(os.path.splitext(f)[0] + '.cochlea')
        print('Saved: ', os.path.splitext(f)[0] + '.cochlea')

        if figure:
            c.make_fig(os.path.splitext(f)[0] + '.pdf')

        c.write_csv(os.path.splitext(f)[0] + '.csv')
        c.render_mask(os.path.splitext(f)[0] + '_mask.tif')
        print('')
        return c


# @graceful_exit('\x1b[1;31;40m' + 'ERROR: Could not load segmentation backend.' + '\x1b[0m')
def _get_backend(unet: bool, cellpose: bool, device: str):
    """ Get the appropriate Backend """
    if unet and not cellpose:
        backend = UNetWatershed(unet_model_path='/media/DataStorage/Dropbox (Partners HealthCare)/'
                                                'HairCellInstance/checkpoints/'
                                                'May11_16-38-12_chris-MS-7C37_checkpoint.hcnet',
                                frcnn_model_path='/media/DataStorage/Dropbox (Partners HealthCare)/HcUnet/TrainedModels/'
                                                 'fasterrcnn_Oct14_06:05.pth',
                                device=device).requires_grad_(False)
    elif not unet and not cellpose:
        backend = SpatialEmbedding(model_loc='/home/chris/Dropbox (Partners HealthCare)/trainHairCellSegmentation/models/May17_10-35-49_CHRISUBUNTU_state_dict.trch',
                                   sigma=torch.tensor([5,5,3]),
                                   scale=torch.tensor([60, 60, 25]),
                                   device=device).requires_grad_(False)
    elif cellpose and not unet:
        from hcat.backends.cellpose import Cellpose  # cellpose announces it makes a log file
        backend = Cellpose(device=device)

    return backend



import torch
import torch.nn as nn
from src.models.depreciated.RDCblock import RDCBlock
from warnings import filterwarnings


filterwarnings("ignore", category=UserWarning)


class RDCNet(nn.Module):
    def __init__(self, in_channels: int, out_channels: int, complexity: int = 20):
        super(RDCNet, self).__init__()

        self.conv3x3_1 = nn.Conv3d(in_channels=in_channels, out_channels=5, padding=1, kernel_size=3)

        self.strided_conv = nn.Conv3d(5, complexity, kernel_size=3, stride=(3,3,2), padding=1)
        self.RDCblock = RDCBlock(in_channels=complexity)
        self.transposed_conv = nn.ConvTranspose3d(in_channels=complexity, out_channels=complexity,
                                                  stride=(3, 3, 2), kernel_size=(3, 3, 3), padding=(1, 1, 1))
        self.out_conv = nn.Conv3d(complexity, out_channels=out_channels, kernel_size=1, padding=0)


        self.batch_norm_out = nn.BatchNorm3d(out_channels)
        self.batch_norm_transpose = nn.BatchNorm3d(complexity)

        self.bn_1 = nn.BatchNorm3d(5)

        self.activation = nn.LeakyReLU()

    def forward(self, x, i: int = 7):
        x = self.activation(self.bn_1(self.conv3x3_1(x)))
        x = self.activation(self.strided_conv(x))
        y = torch.zeros(x.shape).to(x.device)

        for t in range(i):
            in_ = torch.cat((x, y), dim=1)
            y = self.RDCblock(in_) + y

        y = self.activation(self.batch_norm_transpose(self.transposed_conv(y)))
        y = self.out_conv(y)

        return y



import torchvision
import hcat
import os.path
import re
import torch
import wget
from hcat.backends.convNeXt import ConvNeXt

from torchvision.models.detection import FasterRCNN
from torchvision.models.detection.anchor_utils import AnchorGenerator

def init_model():
    backbone = ConvNeXt(in_channels=3, dims=[128, 256, 512, 1024], depths=[3, 3, 27, 3], out_channels=256)
    backbone.out_channels = 256

    anchor_sizes = ((16,), (32,), (64,), (128,))
    aspect_ratios = ((0.5, 1.0, 2.0),) * len(anchor_sizes)
    anchor_generator = AnchorGenerator(sizes=anchor_sizes,
                                       aspect_ratios=aspect_ratios)

    HairCellConvNext = FasterRCNN(backbone,
                                  num_classes=3,
                                  rpn_anchor_generator=anchor_generator,
                                  min_size=256, max_size=600, )

    return HairCellConvNext


def FasterRCNN_from_url(url: str, device: str):
    """ loads model from url """
    path = os.path.join(hcat.__path__[0], 'detection_trained_model.trch')

    # Research Purposes...
    # convnext = '/home/chris/Dropbox (Partners HealthCare)/HairCellInstance/Max_project_detection_resnet_NeXT101_04March_2022_MinValidation.trch'
    # convnext = '/home/chris/Dropbox (Partners HealthCare)/trainHairCellDetection/models/Apr05_09-52-47_CHRISUBUNTU.trch'
    # convnext = '/home/chris/Dropbox (Partners HealthCare)/trainHairCellDetection/models/Apr25_10-08-35_CHRISUBUNTU.trch'
    # convnext = '/home/chris/Dropbox (Partners HealthCare)/trainHairCellDetection/models/Jul15_17-25-42_CHRISUBUNTU.trch'
    # convnext_nocommunitydata = '/media/DataStorage/Dropbox (Partners HealthCare)/HairCellInstance/models/Max_project_detection_resnet_NeXT101_28January_2022_eFinal_just_arturlabtrainingdata.trch'

    if not os.path.exists(path):
        wget.download(url=url, out=path)

    model = init_model()

    print(path)
    checkpoint = torch.load(path, map_location=torch.device('cpu'))
    if 'model_state_dict' in checkpoint:
        checkpoint = checkpoint['model_state_dict']
    model = model.to(device)
    model.load_state_dict(checkpoint)

    return model.eval().to(memory_format=torch.channels_last)


def _is_url(input: str):
    regex = re.compile(
        r'^(?:http|ftp)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)

    # Return true if its a url
    return re.match(regex, input) is not None



import PySimpleGUI as sg
from hcat.lib.cell import Cell
from hcat.lib.cochlea import Cochlea
import skimage.io
import torchvision.utils
from PIL import Image
import io
from hcat.detect import _detect
import torch
from torch import Tensor
import numpy as np
from typing import Union, Optional, List, Dict, Tuple
import hcat.lib.utils
import torch.nn.functional as F
from torchvision.io import encode_png
from hcat.detect import _cell_nms
import matplotlib.pyplot as plt
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg, FigureCanvasAgg
import torchvision.transforms.functional as ttf

MAX_WIDTH, MAX_HEIGHT = 900, 900


# ------------------------------- END OF YOUR MATPLOTLIB CODE -------------------------------

# ------------------------------- Beginning of Matplotlib helper code -----------------------


class gui:
    def __init__(self):
        sg.theme('DarkGrey5')
        plt.ioff()
        # sg.set_options(font='Any')

        button_column = [
            [sg.FileBrowse(size=(15, 1), enable_events=True), ],
            [sg.Button('Load', size=(15, 1)), ],
            [sg.Button('Save', size=(15, 1))],
            [sg.HorizontalSeparator(p=(0, 20))],
            [sg.Text('Cell Diameter (In Pixels)')],
            [sg.Input(size=(10, 1), enable_events=True, default_text=30, key='Diameter'),
             sg.OK(size=(3, 1), key='change_size')],
            [sg.Text('Detection Threshold')],
            [sg.Slider(range=(0, 100), orientation='h', enable_events=True, default_value=80, key='Threshold')],
            [sg.Text('Overlap Threshold')],
            [sg.Slider(range=(0, 100), orientation='h', enable_events=True, default_value=30, key='NMS')],
            [sg.HorizontalSeparator(p=(0, 20))],
            [sg.Button('Run Analysis', size=(15, 1))],
            [sg.Check(text='Live Update: ', key='live_update', enable_events=True)]
        ]
        image_column = [
            [sg.Image(filename='', key='image', size=(900, 900))]
        ]

        adjustment_column = [
            [sg.Canvas(key='Histogram', border_width=2)],
            [sg.Text('Colors to include: '), sg.Check(k='Red', text='Red', default=True, enable_events=True),
             sg.Check(k='Green', text='Green', default=True, enable_events=True),
             sg.Check(k='Blue', text='Blue', default=True, enable_events=True)],
            [sg.Text('', pad=(0, 20))],

            [sg.Text(text='RED', text_color='#ff0000')],
            [sg.Text('Brightness'),
             sg.Slider(key='r_brightness', range=(0, 4), default_value=1, enable_events=True, orientation='h',
                       resolution=0.01, expand_x=True, size=(1, 10))],
            [sg.Text('Contrast  '),
             sg.Slider(key='r_contrast', range=(0, 2), default_value=1, enable_events=True, orientation='h',
                       resolution=0.01,
                       expand_x=True, size=(1, 10))],
            [sg.Text('', pad=(0, 20))],

            [sg.Text(text='GREEN', text_color='#00ff00')],
            [sg.Text('Brightness'),
             sg.Slider(key='g_brightness', range=(0, 4), default_value=1, enable_events=True, orientation='h',
                       resolution=0.01, expand_x=True, size=(1, 10))],
            [sg.Text('Contrast  '),
             sg.Slider(key='g_contrast', range=(0, 2), default_value=1, enable_events=True, orientation='h',
                       resolution=0.01,
                       expand_x=True, size=(1, 10))],

            [sg.Text('', pad=(0, 20))],
            [sg.Text(text='BLUE', text_color='#0000ff')],
            [sg.Text('Brightness'),
             sg.Slider(key='b_brightness', range=(0, 4), default_value=1, enable_events=True, orientation='h',
                       resolution=0.01, expand_x=True, size=(1, 10))],
            [sg.Text('Contrast  '),
             sg.Slider(key='b_contrast', range=(0, 2), default_value=1, enable_events=True, orientation='h',
                       resolution=0.01,
                       expand_x=True, size=(1, 10))],

            [sg.HorizontalSeparator(pad=(0, 30))],
            [sg.Text('OHC: None', key='OHC_count')],
            [sg.Text('IHC: None', key='IHC_count')],

        ]

        layout = [[sg.Column(button_column, vertical_alignment='Top'),
                   sg.VerticalSeparator(),
                   sg.Column(image_column),
                   sg.VerticalSeparator(),
                   sg.Column(adjustment_column, vertical_alignment='Top')
                   ]]

        self.window = sg.Window('HCAT', layout, finalize=True, return_keyboard_events=True)
        print(self.window)

        self.rgb_adjustments = ['r_brightness', 'r_contrast', 'g_brightness', 'g_contrast', 'b_brightness',
                                'b_contrast', ]

        for key in self.rgb_adjustments:
            self.window[key].bind('<ButtonRelease-1>', ' release')

        self.rgb_release = [x + ' release' for x in self.rgb_adjustments]

        # State
        self.__LOADED__ = False
        self.__DIAMETER__ = 30

        # Image Buffers
        self.raw_image = None  # raw image from file
        self.scaled_image = None  # to adjust cell diameter size
        self.scaled_and_adjusted_image = None

        self.display_image_scaled = None
        self.display_image = None
        self.scale_ratio = None

        self.fig_agg = None
        self.fig = None

        self.rgb = None
        self.contrast = None
        self.brightness = None
        self.model = None
        self.device = 'cuda:0' if torch.cuda.is_available() else 'cpu'
        self.cochlea = None

        self.all_boxes = None
        self.all_labels = None
        self.all_scores = None

        self.boxes = None
        self.labels = None
        self.scores = None


    def main_loop(self):

        while True:
            event, values = self.window.read(timeout=20)

            values[
                'Browse'] = '/home/chris/Dropbox (Partners HealthCare)/Gersten et al 2020 Fig 3 images/Fig7_max_projections adjusted/Carbo229_8kHz_63X_Maximum intensity projection.tif'

            self.rgb = [values['Red'], values['Green'], values['Blue']]
            self.contrast = [values['r_contrast'], values['g_contrast'], values['b_contrast']]
            self.brightness = [values['r_brightness'], values['g_brightness'], values['b_brightness']]

            if event == 'Exit' or event == sg.WIN_CLOSED:
                return

            if event == 'live_update' and values['live_update']:
                sg.popup_quick_message('Live model update enabled! May cause crashes on larger images...')

            # Load an image for the first time
            if event == 'Load' and values['Browse'] != '':
                self.load_image(values['Browse'])
                self.draw_image()
                self.render_hist()

                self.__LOADED__ = True

            # Update Histogram:
            if event in self.rgb_adjustments and self.__LOADED__:
                if values['live_update']:
                    self.run_detection_model()
                    self.threshold_and_nms(values['Threshold'], values['NMS'])

                self.draw_image()

            if event in ['Red', 'Green', 'Blue'] + [x + ' release' for x in self.rgb_adjustments] and self.__LOADED__:
                self.render_hist()

            if event == 'change_size' and self.__LOADED__:
                try:
                    self.__DIAMETER__ = float(values['Diameter'])
                except:
                    sg.popup_ok(
                        f'Cannot convert input: {values["Diameter"]} to a number. Defaulting to {self.__DIAMETER__}')

                if values['live_update']:
                    self.run_detection_model()
                    self.threshold_and_nms(values['Threshold'], values['NMS'])

                self.delete_fig_agg()
                self.load_image(values['Browse'])
                self.draw_image()
                self.render_hist()

            if event == 'Run Analysis' and not self.__LOADED__:
                sg.popup_quick_message('No File Loaded')

            if event == 'Run Analysis' and self.__LOADED__:
                self.run_detection_model()
                self.threshold_and_nms(values['Threshold'], values['NMS'])
                self.draw_image()

            if event in ['Threshold', 'NMS']:
                self.threshold_and_nms(values['Threshold'], values['NMS'])
                self.draw_image()

            if event == 'Save' and self.labels is not None:
                self.save(values['Threshold'], values['NMS'], values['Browse'])
                sg.popup_quick_message('Saved!')


    def run_detection_model(self):
        _image = self.scaled_image
        for i in range(3):
            if self.rgb[i]:
                _image[i] = ttf.adjust_brightness(_image[[i], ...], self.brightness[i])
                _image[i] = ttf.adjust_contrast(_image[[i], ...], self.contrast[i])
            else:
                _image[i] = _image[i, ...] * self.rgb[i]

        self.cochlea, self.model = _detect(image_base=_image[0:3, ...],
                                           cell_detection_threshold=0.0,
                                           nms_threshold=1.0,
                                           save_fig=False,
                                           save_png=False,
                                           model=self.model,
                                           return_model=True,
                                           )

        self.all_boxes: Tensor = torch.tensor([cell.boxes.cpu().tolist() for cell in self.cochlea.cells]) * self.ratio
        self.all_scores: Tensor = torch.tensor([cell.scores for cell in self.cochlea.cells])
        self.all_labels: List[str] = [cell.type for cell in self.cochlea.cells]

    @staticmethod
    def image_to_byte_array(image: Image) -> bytes:
        array = io.BytesIO()
        image.save(array, format='PNG')
        return array.getvalue()

    def load_image(self, f: str):

        self.clear_state()

        img: np.array = hcat.lib.utils.load(f, verbose=False)
        scale: int = hcat.lib.utils.get_dtype_offset(img.dtype, img.max())
        self.raw_image: Tensor = hcat.lib.utils.image_to_float(img, scale, verbose=False).to(self.device)
        self.scaled_image: Tensor = hcat.lib.utils.correct_pixel_size_image(self.raw_image, None,
                                                                            cell_diameter=float(self.__DIAMETER__),
                                                                            verbose=False).to(self.device)

        _, x, y = self.scaled_image.shape
        self.ratio = min(900 / x, 900 / y)
        self.display_image_scaled = F.interpolate(self.scaled_image.unsqueeze(0),
                                                  scale_factor=(self.ratio, self.ratio)).squeeze(0)
        print(f'Loaded image and scaled to shape {self.display_image_scaled.shape}')

    def clear_state(self):
        self.all_boxes = None
        self.all_labels = None
        self.all_scores = None

        self.boxes = None
        self.labels = None
        self.scores = None

    def threshold_and_nms(self, thr, nms_thr) -> Tensor:
        cell_key = {'OHC': '#56B4E9', 'IHC': '#E69F00'}

        ind = self.all_scores > (thr / 100)

        _labels = torch.tensor([1 if l == 'OHC' else 2 for l, i in zip(self.all_labels, ind) if i > 0])
        _boxes = self.all_boxes[ind, :]
        _scores = self.all_scores[ind]

        ind = torchvision.ops.nms(_boxes, _scores, nms_thr / 100)  # int indicies

        _boxes = _boxes[ind, :]
        _scores = _scores[ind]

        _labels = ['OHC' if _labels[i] == 1 else 'IHC' for i in ind]

        self.boxes = _boxes
        self.labels = _labels
        self.scores = _scores

        self.update_cell_counts()

    def get_color_histogram(self):
        color = ['r', 'g', 'b']
        fig = plt.figure(figsize=(3, 1), dpi=100)

        fig.patch.set_facecolor('#343434')
        fig.patch.set_alpha(1)

        fig.add_subplot(111)
        ax = plt.gca()

        for i in range(3):
            if self.rgb[i]:
                # hist = torch.histogram(self.display_image[i, ...].float())
                _, _, bars = ax.hist(self.display_image[i, ...].flatten().cpu().numpy(), color=color[i], bins=126,
                        density=True, alpha=0.7)

                max_height = max([b.get_height() for b in bars.patches])
                for b in bars.patches:
                    b.set_height(b.get_height() / max_height)


        plt.ylim([0, 1])
        plt.xlim([0, 255])

        ax.spines.right.set_visible(False)
        ax.spines.left.set_visible(False)
        ax.spines.top.set_visible(False)

        plt.tick_params(top=False, bottom=False, left=False, right=False,
                        labelleft=False, labelbottom=False)

        ax.patch.set_facecolor('#343434')
        plt.tight_layout()

        self.fig = plt.gcf()


    def delete_fig_agg(self):
        if self.fig_agg:
            self.fig_agg.get_tk_widget().forget()
            plt.close('all')
        self.window.refresh()

    def draw_figure(self):
        figure_canvas_agg = FigureCanvasTkAgg(self.fig, self.window['Histogram'].TKCanvas)
        figure_canvas_agg.draw()
        figure_canvas_agg.get_tk_widget().pack(side='top', fill='both', expand=1)
        self.fig_agg = figure_canvas_agg

    def draw_image(self):
        cell_key = {'OHC': '#56B4E9', 'IHC': '#E69F00'}

        _image = self.display_image_scaled.clone()

        for i in range(3):
            if self.rgb[i]:
                _image[i] = ttf.adjust_brightness(_image[[i], ...], self.brightness[i])
                _image[i] = ttf.adjust_contrast(_image[[i], ...], self.contrast[i])

            _image[i] = _image[i, ...] * self.rgb[i]

        _image: Tensor = hcat.lib.utils.make_rgb(_image)
        _image: Tensor = _image.mul(255).round().to(torch.uint8).cpu()

        if self.labels:
            color = [cell_key[l] for l in self.labels]
            _image: Tensor = torchvision.utils.draw_bounding_boxes(image=_image,
                                                                boxes=self.boxes,
                                                                colors=color)

        img = encode_png(_image, 0).numpy().tobytes()
        _, x, y = _image.shape

        self.window['image'].update(data=img, size=(y, x))
        self.display_image = _image

    def render_hist(self):
        self.delete_fig_agg()
        self.get_color_histogram()
        self.draw_figure()

    def save(self, thr, nms, filename):
        _cells = [c for c in self.cochlea.cells if c.scores > thr]
        _cells = self.cell_nms(_cells, nms)
        self.cochlea.cells = _cells

        self.cochlea.write_csv(filename=filename)
        self.cochlea.make_detect_fig(self.scaled_image, filename=filename)

    def update_cell_counts(self):
        if self.labels:
            ihc = sum([l == 'IHC' for l in self.labels])
            ohc = sum([l == 'OHC' for l in self.labels])
        else:
            ihc, ohc = 'None', 'None'

        self.window['IHC_count'].update(f'IHC: {ihc}')
        self.window['OHC_count'].update(f'OHC: {ohc}')


    @staticmethod
    def cell_nms(cells: List[Cell], nms_threshold: float) -> List[Cell]:
        """
        Perforns non maximum supression on the resulting cell predictions

        :param cells: Iterable of cells
        :param nms_threshold: cell iou threshold
        :return: Iterable of cells
        """
        # nms to get rid of cells
        boxes = torch.zeros((len(cells), 4))
        scores = torch.zeros(len(cells))
        for i, c in enumerate(cells):
            boxes[i, :] = c.boxes
            scores[i] = c.scores

        ind = torchvision.ops.nms(boxes, scores, nms_threshold)

        # need to pop off list elements from an int64 tensor
        ind_bool = torch.zeros(len(cells))
        ind_bool[ind] = 1
        for i, val in enumerate(ind_bool):
            if val == 0:
                cells[i] = None

        return [c for c in cells if c]


if __name__ == '__main__':
    gui().main_loop()



from typing import Tuple, Dict, Optional, List
import torch
from torch import Tensor
from torch.autograd import Function
import torch.nn as nn
from typing import Tuple

import triton
import triton.language as tl


@triton.jit
def _embedding_forward_kernel(
        start_ptr,  # *Pointer* to first input vector
        output_ptr,  # *Pointer* to output vector

        # Vector Strides,
        c_stride, x_stride, y_stride, z_stride,

        # Probability Strides,
        n_prob_stride, x_prob_stride, y_prob_stride, z_prob_stride,

        # Centroid Number...
        N,

        # Centroids
        x_center, y_center, z_center,

        # Sigma
        x_sigma, y_sigma, z_sigma,

        # Size of the vector
        n_embed_elements,
        n_output_elements,

        # Constants
        BLOCK_SIZE: tl.constexpr
):
    """

    We want to move over the input tensor of shape [B, C=3, X, Y, Z]
    and do some gaussian operation then put it into an output tensor of shape [B, C=1, X, Y, Z]

    euclidean_norm = sqrt(Δx^2 + Δy^2 + Δz^2) where Δx = (x_embed - x_centroid_i)

                       /    (e_ix - C_kx)^2       (e_iy - C_ky)^2        (e_iz - C_kz)^2   \
    prob_k(e_i) = exp |-1 * ----------------  -  -----------------   -  ------------------  |
                      \     2*sigma_kx ^2         2*sigma_ky ^2          2 * sigma_kz ^2  /

    Returns:

    """
    pid = tl.program_id(axis=0)  # We use a 1D launch grid so axis is 0
    block_start = pid * BLOCK_SIZE

    offsets = block_start + tl.arange(0, BLOCK_SIZE)

    # Create a mask to guard memory operations against out-of-bounds accesses
    embed_mask = offsets < n_embed_elements

    # Load in the vectors into memory
    x = tl.load(start_ptr + offsets, mask=embed_mask)  # Load values of input tensor in memory

    # Initialize intermediary tensors
    probability = tl.zeros(x.shape, dtype=tl.float32)
    sigma = tl.zeros(x.shape, dtype=tl.float32)
    centroid = tl.zeros(x.shape, dtype=tl.float32)

    # We first account for batching!
    c = offsets // c_stride  # Which channel are we in?
    _offsets = offsets - (c * c_stride)

    # Write the X Index
    x_ind = _offsets // x_stride
    _offsets = _offsets - (x_ind * x_stride)

    # Write the Y Index
    y_ind = _offsets // y_stride
    _offsets = _offsets - (y_ind * y_stride)

    # Write the Z Index
    z_ind = _offsets // z_stride

    output_offsets = (x_ind * x_prob_stride) + (y_ind * y_prob_stride) + (z_ind * z_prob_stride) + (N * n_prob_stride)
    output_masks = output_offsets < n_output_elements

    centroid = tl.where(c == 0, x_center, centroid)
    centroid = tl.where(c == 1, y_center, centroid)
    centroid = tl.where(c == 2, z_center, centroid)

    sigma = tl.where(c == 0, x_sigma, sigma)
    sigma = tl.where(c == 1, y_sigma, sigma)
    sigma = tl.where(c == 2, z_sigma, sigma)

    probability = ((x - centroid) * (x - centroid)) / (sigma + 1e-16)

    tl.atomic_add(output_ptr + output_offsets, probability, mask=output_masks)


@triton.jit
def _embedding_backward_kernel(

        previous_grad_ptr,  # [N ,X, Y, Z]  # Iterating over this vector because its the biggest!
        centroid_ptr,  # [N ,3] Use this to figure out which centroid?
        embed_ptr,  # *Pointer* to first input vector # [3, X, Y, Z]
        grad_ptr,  # *Pointer* to output vector # [3, X, Y, Z]

        # Vector Strides,
        n_stride, x_stride, y_stride, z_stride,

        # Centroid Strides
        n_centroid_stride, coord_stride,

        # Sigma
        sigma_ptr,

        # Size of the vector
        embed_numel,
        previous_grad_numel,
        centroid_numel,

        # Constants
        BLOCK_SIZE: tl.constexpr
):
    """
    Effectivly Does This...

    _embed_grad = torch.zeros(ctx.embed.shape, dtype=torch.float32, device=grad_outputs.device)
    sigma = torch.tensor(ctx.sigma, device=grad_outputs.device)
    for n, center in enumerate(ctx.centroids):
        _embed_grad += 2 * (ctx.embed - torch.tensor(center, device=grad_outputs.device).view(3, 1, 1, 1)) / sigma.view(3, 1,1,1) * grad_outputs[[n], ...]
    return _embed_grad, None, None


    """
    pid = tl.program_id(axis=0)  # We use a 1D launch grid so axis is 0
    coord_channel = tl.program_id(axis=1)  # We use a 1D launch grid so axis is 0

    block_start = pid * BLOCK_SIZE
    offsets = block_start + tl.arange(0, BLOCK_SIZE)

    n_ind, x_ind, y_ind, z_ind = get_index(offsets, n_stride, x_stride, y_stride, z_stride)

    # Create a mask to guard memory operations against out-of-bounds accesses
    previous_grad_mask = offsets < previous_grad_numel
    previous_grad = tl.load(previous_grad_ptr + offsets,
                            mask=previous_grad_mask)  # Load values of input tensor in memory

    embed_offsets = (coord_channel * n_stride) + (x_ind * x_stride) + (y_ind * y_stride) + (z_ind * z_stride)
    embed_mask = embed_offsets < embed_numel
    embed = tl.load(embed_ptr + embed_offsets, mask=embed_mask)

    centroid_offsets = (n_ind * n_centroid_stride) + (coord_stride * coord_channel)
    centroid_mask = centroid_offsets < centroid_numel
    center = tl.load(centroid_ptr + centroid_offsets, mask=centroid_mask)

    sigma = tl.load(sigma_ptr + coord_channel)

    grad = 2 * (embed - center) / sigma * previous_grad
    tl.atomic_add(grad_ptr + embed_offsets, grad, mask=embed_mask)


@triton.jit
def get_index(offsets, c_stride, x_stride, y_stride, z_stride):
    # We first account for batching!
    c_ind = offsets // c_stride  # Which channel are we in?
    _offsets = offsets - (c_ind * c_stride)

    # Write the X Index
    x_ind = _offsets // x_stride
    _offsets = _offsets - (x_ind * x_stride)

    # Write the Y Index
    y_ind = _offsets // y_stride
    _offsets = _offsets - (y_ind * y_stride)

    # Write the Z Index
    z_ind = _offsets // z_stride

    return c_ind, x_ind, y_ind, z_ind


class embed2prob3D(Function):
    """
    Performs the vector to Embedding on 4D Inputs!
    """

    @staticmethod
    def forward(ctx, embed: torch.Tensor, sigma: Tuple[float], centroid: Tuple[Tuple[float]]):
        assert embed.ndim == 4
        assert embed.shape[0] == 3

        C, X, Y, Z = embed.shape
        _cs, _xs, _ys, _zs = embed.stride()

        N = len(centroid)

        output = torch.zeros((N, X, Y, Z), device=embed.device, dtype=embed.dtype)

        _cos, _xos, _yos, _zos = output.stride()

        # We need to preallocate the output

        assert embed.is_cuda and output.is_cuda
        assert embed.is_contiguous and output.is_contiguous

        n_embed_elements = embed.numel()
        n_output_elements = output.numel()

        for i, center in enumerate(centroid):
            grid = lambda META: (triton.cdiv(n_embed_elements, META['BLOCK_SIZE']),)

            _embedding_forward_kernel[grid](embed, output,

                                            # Input Strides
                                            c_stride=_cs, x_stride=_xs, y_stride=_ys, z_stride=_zs,

                                            # Output Strides
                                            n_prob_stride=_cos, x_prob_stride=_xos,
                                            y_prob_stride=_yos, z_prob_stride=_zos,

                                            # Which centroid are we?
                                            N=i,
                                            # Centers
                                            x_center=center[0], y_center=center[1], z_center=center[2],

                                            # Sigma
                                            x_sigma=sigma[0], y_sigma=sigma[1], z_sigma=sigma[2],

                                            # Number of Elements
                                            n_embed_elements=n_embed_elements,
                                            n_output_elements=n_output_elements,
                                            BLOCK_SIZE=513)

            torch.cuda.current_stream().synchronize()  # Neccessary to avoid any issues as the cuda stream may be very large!

        ctx.centroids = centroid
        ctx.embed = embed
        ctx.sigma = sigma

        return output

    @staticmethod
    def backward(ctx, grad_outputs: torch.Tensor):
        """
        SUM_{n centroids} = 2 * (vec - x) * sigma * grad_outputs


        # Native pyTorch implementation...
        _embed_grad = torch.zeros(ctx.embed.shape, dtype=torch.float32, device=grad_outputs.device)
        sigma = torch.tensor(ctx.sigma, device=grad_outputs.device)
        for n, center in enumerate(ctx.centroids):
            _embed_grad += 2 * (ctx.embed - torch.tensor(center, device=grad_outputs.device).view(3, 1, 1, 1)) / sigma.view(3, 1,1,1) * grad_outputs[[n], ...]
        return _embed_grad, None, None

        :param ctx:
        :param grad_outputs:
        :return:
        """

        assert grad_outputs.ndim == 4
        assert ctx.embed.ndim == 4
        assert ctx.embed.shape[0] == 3
        assert len(ctx.centroids) == grad_outputs.shape[0]
        assert ctx.embed.shape[1] == grad_outputs.shape[1]
        assert ctx.embed.shape[2] == grad_outputs.shape[2]
        assert ctx.embed.shape[3] == grad_outputs.shape[3]
        assert len(ctx.sigma) == 3

        C, X, Y, Z = ctx.embed.shape
        _cs, _xs, _ys, _zs = ctx.embed.stride()

        output = torch.zeros((3, X, Y, Z), device=grad_outputs.device, dtype=grad_outputs.dtype)

        _cos, _xos, _yos, _zos = output.stride()

        previous_grad_numel = grad_outputs.numel()

        centroids = torch.tensor(ctx.centroids, device=grad_outputs.device)
        sigma = torch.tensor(ctx.sigma, device=grad_outputs.device)
        n_centroid_stride, coord_stride = centroids.stride()

        grid = lambda META: (triton.cdiv(previous_grad_numel, META['BLOCK_SIZE']), 3)  # 2D Lauch Grid!!!

        _embedding_backward_kernel[grid](
            previous_grad_ptr=grad_outputs,  # [N ,X, Y, Z]  # Iterating over this vector because its the biggest!
            centroid_ptr=centroids,  # [N ,3] Use this to figure out which centroid?
            embed_ptr=ctx.embed,  # *Pointer* to first input vector # [3, X, Y, Z]
            grad_ptr=output,  # *Pointer* to output vector # [3, X, Y, Z]

            # Vector Strides,
            n_stride=_cs, x_stride=_xs, y_stride=_ys, z_stride=_zs,

            # Centroid Strides
            n_centroid_stride=n_centroid_stride, coord_stride=coord_stride,

            # Sigma
            sigma_ptr=sigma,

            # Size of the vector
            embed_numel=ctx.embed.numel(),
            previous_grad_numel=grad_outputs.numel(),
            centroid_numel=centroids.numel(),

            # Constants
            BLOCK_SIZE=512)

        torch.cuda.current_stream().synchronize()

        return output, None, None


embed2prob = embed2prob3D.apply


class EmbeddingToProbability(nn.Module):
    def __init__(self):
        super(EmbeddingToProbability, self).__init__()

    def forward(self, embedding, centroids, sigma):

        if embedding.is_cuda():
            return self._forward_triton(embedding, centroids, sigma)
        else:
            return self._forward_torch(embedding, centroids, sigma)

    @torch.jit.ignore
    def _forward_triton(self, embedding, centroids, sigma):
        return torch.exp(embed2prob(embedding, centroids, sigma))

    def _forward_torch(self, embedding: Tensor, centroids: Tensor, sigma: Tensor) -> Tensor:
        """
        Calculates the euclidean distance between the centroid and the embedding
        embedding [B, 3, X, Y, Z] -> euclidean_norm[B, 1, X, Y, Z]
        euclidean_norm = sqrt(Δx^2 + Δy^2 + Δz^2) where Δx = (x_embed - x_centroid_i)

                             /    (e_ix - C_kx)^2       (e_iy - C_ky)^2        (e_iz - C_kz)^2   \
          prob_k(e_i) = exp |-1 * ----------------  -  -----------------   -  ------------------  |
                            \     2*sigma_kx ^2         2*sigma_ky ^2          2 * sigma_kz ^2  /


        Example:


        :param embedding: [B, K=3, X, Y, Z] embedding tensor where K is the likely centroid component: {X, Y, Z}
        :param centroids: [B, N, K_true=3] object centroids where N is the number of instances in the image and K_true is centroid {x, y, z}
        :param sigma: Tensor of shape = (1) or (embedding.shape)
        :return: [B, N, X, Y, Z] of probabilities for instance N
        """
        b, _, x, y, z = embedding.shape
        _, n, _ = centroids.shape

        # Centroids might be empty. In this case, return empty array!
        if n == 0:
            return torch.zeros((b, n, x, y, z), device=embedding.device)

        sigma = sigma + torch.tensor([1e-16], device=centroids.device)  # when sigma goes to zero, things tend to break

        if sigma.numel() == 1:
            sigma = torch.cat((sigma, sigma, sigma), dim=0)

        b, _, x, y, z = embedding.shape
        _, n, _ = centroids.shape
        # prob = torch.zeros((b, n, x, y, z), device=embedding.device)
        prob_list = torch.jit.annotate(List[Tensor], [])

        # Common operation. Done outside of loop for speed.
        newshape = (centroids.shape[0], 3, 1, 1, 1)
        sigma = sigma.pow(2).mul(2).view(newshape).mul(-1)

        # Calculate euclidean distance between centroid and embedding for each pixel and
        # turn that distance to probability and put it in preallocated matrix for each n
        # In eval mode uses in place operations to save memory!

        # A bit scuffed but should optimize well in torchscript
        for i in range(n):
            prob_list += [torch.exp(
                (embedding - centroids[:, i, :].view(newshape)).pow(2)
                    .div(sigma)
                    .sum(dim=1)
            ).squeeze(1)]
        prob = torch.stack(prob_list, dim=1)

        return prob




import warnings

import skimage.scripts.skivi
import matplotlib.pyplot as plt
import torch
from torch import Tensor
from torch.utils.data import DataLoader
import torchvision.transforms.functional as TF
import numpy as np
import glob
import os.path
import shutil
from tqdm import tqdm
import skimage.io as io
from typing import Dict, List, Union
from skimage.morphology import ball
import xml
from torch.utils.data import Dataset
import hcat.lib.utils
from typing import Callable

import hcat.train.transforms as t


class dataset(DataLoader):
    def __init__(self, path: Union[List[str], str], transforms=None, random_ind: bool = False, min_cell_volume: int = 4000,
                 duplicate_poor_data: bool = False):
        super(DataLoader, self).__init__()

        # Functions to correct for Bad Masks
        self.colormask_to_mask = t.colormask_to_mask()
        self.transformation_correction = t.transformation_correction(min_cell_volume)  # removes bad masks

        # Reassigning variables
        self.mask = []
        self.image = []
        self.centroids = []
        self.transforms = transforms
        self.dpd = duplicate_poor_data

        path: List[str] = [path] if isinstance(path, str) else path
        print('WARNING: Memory PINNED!!!')

        # Find only files
        for p in path:
            self.files = glob.glob(os.path.join(p, '*.labels.tif'))
            for f in self.files:
                image_path = os.path.splitext(f)[0]
                image_path = os.path.splitext(image_path)[0] + '.tif'

                if image_path.find(
                        '-DUPLICATE') > 0 and not self.dpd:  # If -DUPLICATE flag is there and we want extra copies continue
                    continue

                image = torch.from_numpy(io.imread(image_path).astype(np.uint16) / 2 ** 16).unsqueeze(0)
                image = image.transpose(1, 3).transpose(0, -1).squeeze()[[0, 2, 3], ...]

                mask = torch.from_numpy(io.imread(f)).transpose(0, 2).unsqueeze(0)

                self.mask.append(mask.float())
                self.image.append(image.float())
                self.centroids.append(torch.tensor([0, 0, 0]))

            # implement random permutations of the indexing
            self.random_ind = random_ind

            if self.random_ind:
                self.index = torch.randperm(len(self.mask))
            else:
                self.index = torch.arange((len(self.mask)))

    def __len__(self) -> int:
        return len(self.mask)

    def __getitem__(self, item: int) -> Dict[str, torch.Tensor]:

        with torch.no_grad():
            item = self.index[item]
            data_dict = {'image': self.image[item], 'masks': self.mask[item], 'centroids': self.centroids[item]}
            did_we_get_an_output = False

            while not did_we_get_an_output:
                try:
                    if self.transforms is not None:
                        data_dict = self.transforms(data_dict)
                        did_we_get_an_output = True
                except RuntimeError:
                    continue

            data_dict = self.colormask_to_mask(data_dict)
            assert torch.isnan(data_dict['centroids']).sum() == 0
            data_dict = self.transformation_correction(data_dict)  # removes bad masks

        return data_dict

    def step(self) -> None:
        if self.random_ind:
            self.index = torch.randperm(len(self.mask))


class nul_dataset(DataLoader):
    def __init__(self, path, transforms=None, random_ind: bool = False):

        super(DataLoader, self).__init__()

        # Find only files
        files = glob.glob(os.path.join(path, '*.tif'))

        self.image = []
        self.centroids = []
        self.transforms = transforms

        for f in files:
            image = torch.from_numpy(io.imread(f).astype(np.uint16) / 2 ** 16).unsqueeze(0)
            image = image.transpose(1, 3).transpose(0, -1).squeeze()[[0, 2, 3], ...]
            self.centroids.append(torch.tensor([0, 0, 0]))
            self.image.append(image.float())

        # implement random permutations of the indexing
        self.random_ind = random_ind

        if self.random_ind:
            self.index = torch.randperm(len(self.image))
        else:
            self.index = torch.arange((len(self.image)))

    def __len__(self) -> int:
        return len(self.mask)

    def __getitem__(self, item: int) -> Dict[str, torch.Tensor]:

        item = self.index[item]
        shape = self.image[item].shape
        shape[0] = 1  # override the channel shape (4 -> 1)

        data_dict = {'image': self.image[item], 'masks': torch.zeros(shape), 'centroids': self.centroids[item]}
        did_we_get_an_output = False

        while not did_we_get_an_output:
            try:
                if self.transforms is not None:
                    data_dict = self.transforms(data_dict)
                    did_we_get_an_output = True
            except RuntimeError:
                continue

        return data_dict

    def step(self) -> None:
        if self.random_ind:
            self.index = torch.randperm(len(self.image))


class em_dataset(DataLoader):
    def __init__(self, path, transforms=None, random_ind: bool = False, big_path =None):
        super(DataLoader, self).__init__()

        # Find only files
        files = glob.glob(os.path.join(path, '*.labels.tif'))
        print(files)

        self.mask = []
        self.image = []
        self.centroids = []
        self.transforms = transforms

        # Functions to correct for Bad Masks
        self.colormask_to_mask = t.colormask_to_mask()
        self.transformation_correction = t.transformation_correction(4000)  # removes bad masks


        for f in files:
            image_path = os.path.splitext(f)[0]
            image_path = os.path.splitext(image_path)[0] + '.tif'
            image = torch.from_numpy(io.imread(image_path).astype(np.uint16) / 2 ** 8).unsqueeze(-1)
            if image.shape[-1] > 1:
                image = image.transpose(0, -1).transpose(1,2).squeeze()[[0, 2, 3], ...]
            else:
                image = image.transpose(0, -1).transpose(1,2).squeeze().unsqueeze(0)

            mask = torch.from_numpy(io.imread(f)).transpose(0, 2).unsqueeze(0)

            self.mask.append(mask.float())
            self.image.append(image.float())
            self.centroids.append(torch.tensor([0]))


        if big_path:
            if not isinstance(big_path, list):
                big_path = [big_path]

            for path in big_path:
                stacks = glob.glob(path+'/mito_train/*.tif')
                imgs = glob.glob(path+'/im/*.png')
                stacks.sort()
                imgs.sort()

                big_img = torch.zeros((4096, 4096, len(stacks)), dtype=torch.uint8)
                big_mask = torch.zeros(big_img.shape, dtype=torch.int16)
                for i, s in tqdm(enumerate(stacks), total=len(stacks)):
                    m = io.imread(s)
                    big_mask[:,:,i] = torch.from_numpy(m.astype(np.int32)).type(torch.int16).mul_(-1).add_(2**16)

                    m = io.imread(imgs[i])
                    big_img[:,:,i]  = torch.from_numpy(m.astype(np.int32)).type(torch.int16)

                self.mask.append(big_mask.float().unsqueeze(0))
                self.image.append(big_img.float().unsqueeze(0))
                self.centroids.append(torch.tensor([0]))

                del big_mask, big_img

        # implement random permutations of the indexing
        self.random_ind = random_ind

        if self.random_ind:
            self.index = torch.randperm(len(self.mask))
        else:
            self.index = torch.arange((len(self.mask)))

    def __len__(self) -> int:
        return len(self.mask)


    def step(self) -> None:
        if self.random_ind:
            self.index = torch.randperm(len(self.mask))

    def __getitem__(self, item: int) -> Dict[str, torch.Tensor]:
        with torch.no_grad():
            item = self.index[item]
            data_dict = {'image': self.image[item], 'masks': self.mask[item], 'centroids': self.centroids[item]}
            did_we_get_an_output = False
            i = 0
            while not did_we_get_an_output:
                i += 1
                if self.transforms is not None:
                    data_dict = self.transforms(data_dict)
                    did_we_get_an_output = True

            data_dict = self.colormask_to_mask(data_dict)
            assert torch.isnan(data_dict['centroids']).sum() == 0
            data_dict = self.transformation_correction(data_dict)  # removes bad masks


        return data_dict


class synthetic:
    def __init__(self):
        pass

    def __len__(self):
        return 1

    def __getitem__(self, item: int) -> Dict[str, torch.Tensor]:
        ball = torch.from_numpy(skimage.morphology.ball(15))
        mask = torch.zeros((1, 156, 156, 31))
        image = torch.zeros((4, 156, 156, 31))

        x = torch.randint(156 - 32, (1,))
        y = torch.randint(156 - 32, (1,))

        mask[0, x:x + 31, y:y + 31, :] = ball
        image[1, x:x + 31, y:y + 31, :] = ball
        image = (2 * image / 3) + (torch.rand((4, 156, 156, 31)) / 3)
        centroids = torch.zeros((1, 3))
        centroids[0, :] = torch.tensor([x + 31 // 2, y + 31 // 2, 31 // 2])

        return {'masks': mask.cuda(), 'image': image.cuda(), 'centroids': centroids.cuda()}

    def step(self):
        pass


class detection_dataset(Dataset):
    def __init__(self, path: Union[List[str], str], transforms, random_ind: bool = False, min_cell_volume: int = 20*10,
                 duplicate_poor_data: bool = False, simple_class: bool = True, batch_size: int = 1, pad_size: int = 200):
        super(Dataset, self).__init__()

        # Functions to correct for Bad Masks
        self.transformation_correction = t.transformation_correction(min_cell_volume)  # removes bad masks

        if transforms is None: raise RuntimeError('Transformations Necessary for data loading.')

        # Reassigning variables
        self.mask = []
        self.image = []
        self.centroids = []
        self.boxes = []
        self.labels = []
        self.files = []
        self.transforms = transforms
        self.dpd = duplicate_poor_data
        self.simple_class = simple_class
        self.pad_size = pad_size
        self.device = 'cuda:0'

        path: List[str] = [path] if isinstance(path, str) else path

        # Find only files
        for p in path:
            self.files.extend(glob.glob(f'{p}{os.sep}*.xml'))

        print('WARNING: Image hacks in src.train.dataloader.detection_dataset')
        for f in self.files:
            image_path = f[:-4:]+ '.tif'

            if image_path.find(
                    '-DUPLICATE') > 0 and not self.dpd:  # If -DUPLICATE flag is there and we want extra copies continue
                continue

            # [c, X, Y] -> [c, x, y, z=1]
            image: np.array = io.imread(image_path)
            scale: int = hcat.lib.utils.get_dtype_offset(image.dtype)
            image: Tensor = TF.pad(torch.from_numpy(image / scale), self.pad_size).unsqueeze(-1)

            if image.max() < 0.2:
                image = image.div(image.max()).mul(0.8)
            if image[1, ...].max() < 0.3:
                image[1, ...] = image[1, ...].mul(2)

            mask = torch.empty((0,0,0,0))

            self.mask.append(mask)
            self.image.append(image.float().to(self.device, memory_format=torch.channels_last))
            self.centroids.append(torch.tensor([0, 0, 0]))

            tree = xml.etree.ElementTree.parse(f)
            root = tree.getroot()
            bbox_loc = []
            class_labels = []

            for c in root.iter('object'):
                for cls in c.iter('name'):
                    class_labels.append(cls.text)

                for a in c.iter('bndbox'):
                    x1 = int(a.find('xmin').text)
                    y1 = int(a.find('ymin').text)
                    x2 = int(a.find('xmax').text)
                    y2 = int(a.find('ymax').text)
                    bbox_loc.append([x1, y1, x2, y2])

            for i, s in enumerate(class_labels):
                if s == 'OHC':
                    class_labels[i] = 1
                elif s == 'OHC1':
                    class_labels[i] = 1
                elif s == 'OHC2':
                    class_labels[i] = 2
                elif s == 'OHC3':
                    class_labels[i] = 3
                elif s == 'IHC':
                    class_labels[i] = 4
                else:
                    class_labels[i] = 999
                    print('Unidentified Label in XML file of ' + f, '\n Assigning Value 999')

            class_labels = torch.tensor(class_labels)

            # For testing - reduce row specific labels in cell specific (IHC or OHC)
            if self.simple_class:
                class_labels[class_labels == 2] = 1  # OHC2 -> OHC
                class_labels[class_labels == 3] = 1  # OHC3 -> OHC
                class_labels[class_labels == 4] = 2  # IHC

            b = torch.tensor(bbox_loc).to(self.device) + self.pad_size
            c = class_labels.to(self.device)
            b = b[c < 10, :]
            c = c[c < 10]

            # ind = b[:, 0:2] < b[:, 2:]
            # ind = torch.logical_and(ind[:,0], ind[:,1])
            #
            # b = b[ind, :]
            # c = c[ind]

            self.boxes.append(b)
            self.labels.append(c)

        print(f'TOTAL DATASET SIZE: {len(self.files)}')

        # implement random permutations of the indexing
        self.random_ind = random_ind
        self.batch_size = batch_size

        self.index = self._batch_ind(batch_size=self.batch_size, dataset_len=len(self.image), random_ind=self.random_ind)

    @staticmethod
    def _batch_ind( batch_size: int, dataset_len: int, random_ind: bool = False) -> List[List[int]]:
        """

        :param batch_size:
        :param dataset_len:
        :return:
        """

        end = dataset_len + batch_size - (dataset_len % batch_size)
        base_ind: Tensor = torch.arange(0, end) # should make evenly

        if not base_ind.shape[0] % batch_size == 0:
            print(base_ind.shape, batch_size, dataset_len, dataset_len % batch_size, base_ind.shape[0] % batch_size)
            raise ValueError

        indicies = []
        _ind = []
        for i, ind in enumerate(base_ind):
            if ind >= dataset_len:
                base_ind[i] -= dataset_len

        # Shuffle the indicies if we do a random permutation
        if random_ind: base_ind = base_ind[torch.randperm(base_ind.shape[0])]

        for i, ind in enumerate(base_ind):
            _ind.append(base_ind[i].item())
            if len(_ind) == batch_size:
                indicies.append(_ind)
                _ind = []

        return indicies

    def __len__(self) -> int:
        return len(self.index)

    def __getitem__(self, item: int) -> Union[Dict[str, torch.Tensor], Dict[str, Union[List[torch.Tensor], torch.Tensor]]]:

        out = []

        with torch.no_grad():
            item_list: List[int] = self.index[item]

            for index, item in enumerate(item_list):
                data_dict = {'image': self.image[item],
                             'masks': self.mask[item],
                             'centroids': self.centroids[item],
                             'boxes': self.boxes[item],
                             'labels': self.labels[item]}

                # Transformation pipeline
                data_dict = self.transforms(data_dict) # Apply transforms

                data_dict['boxes'] = torch.round(data_dict['boxes'].float())
                c, x, y, z = data_dict['image'].shape
                data_dict['image'] = torch.cat((torch.zeros((1, x, y, z), device=data_dict['image'].device),
                                                data_dict['image']), dim=0)[:, :, :, 0]
                data_dict.pop('masks', torch.tensor([]))

                for key in data_dict:
                    data_dict[key].detach()

                out.append(data_dict)

            return out  # List[Dict[str, Tensor]]

    def step(self) -> None:
        if self.random_ind:
            self.index = self._batch_ind(batch_size=self.batch_size, dataset_len=len(self.image), random_ind=self.random_ind)
        else:
            warnings.warn('Stepping when random_ind is set to false has no effect.')

    def cuda(self):
        self.image = [x.cuda() for x in self.image]
        self.mask = [x.cuda() for x in self.mask]
        self.centroids = [x.cuda() for x in self.centroids]
        self.boxes = [x.cuda() for x in self.boxes]
        self.labels = [x.cuda() for x in self.labels]

        return self

@torch.jit.script
def colormask_to_torch_mask(colormask: torch.Tensor) -> torch.Tensor:
    """

    :param colormask: [C=1, X, Y, Z]
    :return:
    """
    uni = torch.unique(colormask)
    uni = uni[uni != 0]
    num_cells = len(uni)

    c, x, y, z = colormask.shape

    shape = (num_cells, x, y, z)
    mask = torch.zeros(shape)

    for i, u in enumerate(uni):
        mask[i, :, :, :] = (colormask[0, :, :, :] == u)

    return mask


@torch.jit.script
def colormask_to_centroids(colormask: torch.Tensor) -> torch.Tensor:
    uni = torch.unique(colormask)
    uni = uni[uni != 0]
    num_cells = len(uni)  # cells are denoted by integers 1->max_cell
    shape = (num_cells, 3)
    centroid = torch.zeros(shape)

    for i, u in enumerate(uni):
        indexes = torch.nonzero(colormask[0, :, :, :] == u).float()
        centroid[i, :] = torch.mean(indexes, dim=0)

    # centroid[:, 0] /= colormask.shape[1]
    # centroid[:, 1] /= colormask.shape[2]
    # centroid[:, 2] /= colormask.shape[3]

    return centroid



import torch
import torch.nn.modules.batchnorm
import hcat.lib.utils

@torch.no_grad()
def update_bn(loader, model, device=None):
    r"""Updates BatchNorm running_mean, running_var buffers in the model.

    It performs one pass over data in `loader` to estimate the activation
    statistics for BatchNorm layers in the model.
    Args:
        loader (torch.utils.data.DataLoader): dataset loader to compute the
            activation statistics on. Each data batch should be either a
            tensor, or a list/tuple whose first element is a tensor
            containing data.
        model (torch.nn.Module): model for which we seek to update BatchNorm
            statistics.
        device (torch.device, optional): If set, data will be transferred to
            :attr:`device` before being passed into :attr:`model`.

    Example:
        >>> loader, model = ...
        >>> torch.optim.swa_utils.update_bn(loader, model)

    .. note::
        The `update_bn` utility assumes that each data batch in :attr:`loader`
        is either a tensor or a list or tuple of tensors; in the latter case it
        is assumed that :meth:`model.forward()` should be called on the first
        element of the list or tuple corresponding to the data batch.
    """
    momenta = {}
    for module in model.modules():
        if isinstance(module, torch.nn.modules.batchnorm._BatchNorm):
            module.running_mean = torch.zeros_like(module.running_mean)
            module.running_var = torch.ones_like(module.running_var)
            momenta[module] = module.momentum

    if not momenta:
        return

    was_training = model.training
    model.train()
    for module in momenta.keys():
        module.momentum = None
        module.num_batches_tracked *= 0

    for batch_ind, data_dict in enumerate(loader):
        images, data_dict = hcat.lib.utils.prep_dict(data_dict, device)  # Preps output to handle potential batched data
        output = model(images, data_dict)

    # for input in loader:
    #     if isinstance(input, (list, tuple)):
    #         input = input[0]
    #     if device is not None:
    #         input = input.to(device)
    #
    #     model(input)

    for bn_module in momenta.keys():
        bn_module.momentum = momenta[bn_module]
    model.train(was_training)




from __future__ import annotations

import torch
from torch import Tensor
import time
import matplotlib.pyplot as plt
from typing import List, Union, Optional, Iterable, Dict
from hcat.lib.cell import Cell
from hcat.lib.utils import graceful_exit
from io import BytesIO
from numpy import int8

import lz4.frame as compalg

from xml.etree import ElementTree
import glob
import os.path
import re
from skimage.io import imsave
import skimage.segmentation


class Cochlea:
    def __init__(self,
                 mask: Tensor = None,
                 start_time: str = None,
                 analysis_time: str = None,
                 analysis_type: str = None,  # Either 'detect' or 'segment'
                 filename: str = '',
                 path: str = '',
                 script_dir: str = None,
                 leica_metadata: ElementTree = None,
                 im_shape: Union[Tensor, tuple] = None,
                 cochlear_distance: Tensor = None,
                 cochlear_length: Tensor = None,
                 curvature: Tensor = None,
                 apex: Tensor = None,
                 cells: List[Cell] = None):
        """
        Dataclass of whole cochlear results.

        Useful for both storage, preliminary analysis, and processing of results.

        :cvar *float* gain: gain of detector at a particular channel
        :cvar *float* laser: laser intensity of a particular channel
        :cvar *str* litter: animal litter identification number
        :cvar *str* animal_id: animal identification number
        :cvar *str* analysis_time: total time of analysis
        :cvar *str* start_time: start time of analysis
        :cvar *str* filename: filename of image being analyzed
        :cvar *str* path: filepath of image being analyzed
        :cvar Iterable[float] voxel_size: voxel size (nm)
        :cvar Iterable[int] im_shape: shape of analyzed image
        :cvar List[Cell] cells: List of detected cell objects
        :cvar int num_cells: total number of detected objects
        :cvar Tensor mask: predicted instance segmentation mask
        :cvar Tensor curvature: 2D array of predicted cochlear curvature
        :cvar Tensor cochlear_length: length of cochlea, inferred from curvature
        :cvar Tensor cochlear_percent: percent length of cochlea, inferred from curvature
        :cvar Tensor apex: estimated apex of cochelea
        :cvar str script_dir: directory of analysis scripts used for this cochlea
        :cvar Dict[str,str] scripts: dictionary containing all analysis scripts

        :param mask: whole cochlea predicted segmentation mask
        :param start_time: start time of analysis
        :param analysis_time: total analysis time (seconds)
        :param filename: filename of analyzed image
        :param path: path to analyzed file
        :param script_dir: directory containing analysis scripts
        :param leica_metadata: parsed xml ElementTree from lif file
        :param im_shape: shape of analyzed image
        :param cochlear_percent: predicted array of cochlear percentage corresponding to curvature array
        :param cochlear_length: predicted cochlear length in mm
        :param curvature: predicted cochlear curvature
        :param apex: predicted apex location
        :param cells: list of each predicted cell object
        """

        self.analysis_time = analysis_time
        self.start_time = start_time
        self.analysis_date = time.asctime()
        self.analysis_type = analysis_type
        self.filename = filename
        self.path = path

        # Image params from filename (Semiprivate)
        self._gain = None
        self._laser = None
        self._promoter = None
        self._litter = None
        self._animal_id = None

        # Inferred from filename
        # Setters perform regex on filename and assign appropriate value
        # If no match is found, sets val at None
        self.gain: str = filename
        self.laser: str = filename
        self.promoter: str = filename  # Not Implemented!!!
        self.litter: str = filename
        self.animal_id: str = filename

        self.leica_metadata = leica_metadata

        # Infered from metadata (semiprivate)
        # NOT IMPLEMENTED!!!
        self._x_pix_size = None
        self._y_pix_size = None
        self._z_pix_size = None

        # Infered from metadata
        # NOT IMPLEMENTED!!!
        self.x_pix_size = None
        self.y_pix_size = None
        self.z_pix_size = None

        # Cochlear Analysis
        self.im_shape = im_shape
        self.cells: List[Cell] = cells if cells is not None else []
        self.num_cells = len(self.cells)
        self.mask = mask
        self.curvature = curvature
        self.cochlear_length = cochlear_length
        self.cochlear_distance = cochlear_distance
        self.apex = apex

        # gathered at runtime
        self.script_dir = os.getcwd() if script_dir is None else script_dir
        self.scripts = self._get_python_scripts(self.script_dir)

    ####################################################################################################################
    #                                                Getters and Setters                                               #
    ####################################################################################################################

    @staticmethod
    def _get_python_scripts(script_dir):
        python_files = {}
        script_dir = os.path.join(script_dir, '')
        python_files_list = glob.glob(script_dir + '**/*.py', recursive=True)
        ipy_files_list = glob.glob(script_dir + '**/*.ipynb', recursive=True)

        for f in python_files_list:
            file = open(f, 'r')
            python_files[f] = file.read()
            file.close()

        for f in ipy_files_list:
            file = open(f, 'r')
            python_files[f] = file.read()
            file.close()

        return python_files

    @property
    def animal_id(self):
        return self._litter

    @animal_id.setter
    def animal_id(self, filename):
        match = re.search(' m\d ', filename) if filename else None
        self._animal_id = match[0] if match is not None else None

    @property
    def litter(self):
        return self._litter

    @litter.setter
    def litter(self, filename):
        test_string = 'Oct 22 AAV2-PHP.B-CMV Olga L17 m1 G200 L0.25.lif'
        match = re.search(' L\d\d?\d?', filename) if filename else None
        self._litter = match[0] if match is not None else None

    @property
    def gain(self):
        return self._gain

    @gain.setter
    def gain(self, filename):
        match = re.search('G\d\d?\d?', filename) if filename else None
        self._gain = match[0] if match is not None else None

    @property
    def laser(self):
        return self.laser

    @laser.setter
    def laser(self, filename):
        match = re.search('L0.\d?\d?\d?', filename) if filename else None
        self._laser = match[0] if match is not None else None

    ####################################################################################################################
    #                                                Compression and Saving                                            #
    ####################################################################################################################

    def save(self, filename: str, compress: Optional[bool] = False) -> None:
        """
        Meant to be invoked after class is full of data as a compressed bit object from torch.save

        :param filename: path and filename by which to save
        :param compress: should object be compressed? Default is False. Mask is compressed automatically!
        :return: None
        """
        if compress:
            self._save(self, filename)
        else:
            torch.save(self, filename)

    @classmethod
    def load(cls, filename: str) -> Cochlea:
        """
        Instantiate new class instance from a saved file.

        :param filename: path to file
        :return: Cochlea object
        """
        try:
            return cls._load(filename)
        except Exception:  # Throws _pickle.UnpicklingError if uncompressed
            return torch.load(filename)

    @staticmethod
    def _save(obj, filename: str):
        """
        Saves any object in a compressed file format

        :param obj: python object
        :return: None
        """
        f = BytesIO()
        torch.save(obj, f)
        file = open(filename, 'wb')
        out = compalg.compress(f.getbuffer())
        file.write(out)
        file.close()

    @staticmethod
    def _load(filename):
        file = compalg.open(filename, 'rb')
        return torch.load(file)

    def compress_mask(self, vebose: bool = False) -> None:
        """
        Compressed just the segmentation mask, which is usually the largest object in the class. This is usefull as it
        may be advantageous to load the class quickly without accessing the mask - for instance, getting metadata and
        nothing more.

        :param vebose: print compression statistics.
        :return: None
        """
        if isinstance(self.mask, Tensor):
            self.mask = self._compress_mask(self.mask, vebose)
        else:
            raise ValueError(f'Cannot compress mask of type {type(self.mask)}.')

    def decompress_mask(self) -> None:
        """
        Inverse operation to self.compress_mask. Will decompress self.mask from bytes to Tensor using the
        lz4 algorithm.

        :return: None
        """
        if isinstance(self.mask, bytes):
            self.mask = self._decompress_mask(self.mask)
        elif isinstance(self.mask, Tensor):
            raise ValueError(f'Cannot decompress mask of tpye {type(self.mask)}.')

    @staticmethod
    def _compress_mask(mask: Tensor, verbose: bool) -> bytes:
        """ compressed mask via lz4 algorithm """
        start = time.clock_gettime_ns(0)
        shape = mask.shape
        f = BytesIO()
        torch.save(mask, f)
        l = len(f.getbuffer())
        mask = compalg.compress(f.getbuffer())
        end = time.clock_gettime_ns(0)
        f.close()

        if verbose:
            print(f'Compressed mask with compression ratio of {l / len(mask)} in {(end - start) / 1e9} seconds')

        return mask

    @staticmethod
    def _decompress_mask(byte_mask: bytes) -> Tensor:
        return torch.load(BytesIO(compalg.decompress(byte_mask)))

    ####################################################################################################################
    #                                                    Analysis                                                      #
    ####################################################################################################################

    def render_mask(self, path: str,
                    random_color: Optional[bool] = False,
                    as_single_z_frames: Optional[Union[bool, str]] = False,
                    outline: Optional[bool] = False) -> None:
        """
        Renders tiff image output of predicted segmentation mask and saves to file.
        Can either render the volume mask with each pixel being a cell id, or a random color.

        :param path: path (including filename and extension) where to save the render
        :param random_color: if True, renders each cell with a unique color
        :param as_single_z_frames: NotImplemented
        :param outline: render only cell outlines instead of whole volume
        :return: None
        """

        was_compressed = False
        if isinstance(self.mask, bytes):
            self.decompress_mask()
            was_compressed = True
        c, x, y, z = self.mask.shape

        if outline:
            ind = torch.from_numpy(skimage.segmentation.find_boundaries(self.mask)).gt(0)
        else:
            ind = torch.ones(self.mask.shape).gt(0)

        # Rust speedup
        if random_color and not as_single_z_frames:
            colored_mask = torch.zeros((3, x, y, z), dtype=torch.int32)
            for i in range(3):
                colored_mask[i, ...][ind[0, ...]] = self.mask[ind][0, ...]

            print('Rendering Multicolored Tiff, may take a while...')
            t1 = time.clock_gettime_ns(0)
            colored_mask = self._render(colored_mask, self.num_cells)
            t2 = time.clock_gettime_ns(0)
            print(f'took {(t2 - t1) / 1e9} seconds')
            imsave(path, colored_mask.squeeze(0).numpy().transpose(3, 1, 2, 0, ))
        elif not random_color and not as_single_z_frames:
            imsave(path, self.mask.squeeze(0).int().numpy().astype(int8).transpose(2, 0, 1))
        # elif as_single_z_frames:
        #     os.mkdir(path)
        #     for i in range(z):

        if was_compressed:
            self.compress_mask()

    @staticmethod
    @torch.jit.script
    def _render(mask: Tensor, numcells: int) -> Tensor:
        for i in range(numcells):
            torch.random.manual_seed(i)
            color = torch.randint(low=0, high=255, size=(3,))
            index = mask == (i + 1)
            for i in range(3):
                mask[i, ...][index[0, ...]] = int(color[i])
        return mask

    # @graceful_exit('\x1b[1;31;40m' + 'ERROR: csv generation failed.' + '\x1b[0m')
    def write_csv(self, filename: Optional[Union[bool, str]] = None) -> None:
        """
        Write results of cochlea object to a csv file for futher statistical analysis.

        .. warning:
           Will not raise an error upon failure, instead returns None and prints to standard out

        :param filename: filename to save csv as. If unset, uses image filename.
        :return: None
        """
        if self.analysis_type == 'segment':
            label = 'cellID,frequency,percent_loc,x_loc,y_loc,z_loc,volume,summed,'
            for c in ['myo', 'dapi', 'actin', 'gfp']:
                label += f'{c}_mean,{c}_median,{c}_std,{c}_var,{c}_min,{c}_max,{c}_%zero,{c}_%saturated,'

            # print(filename)

            if filename is None and self.filename is not None:
                filename = os.path.splitext(self.filename)[0] + '.csv'  # Remove .lif and add .csv
            elif filename is None and self.filename is None:
                filename = 'analysis.csv'

            f = open(filename, 'w')
            f.write(f'Filename: {self.filename}\n')
            f.write(f'Analysis Date: {self.analysis_date}\n')
            f.write(label[:-1:] + '\n')  # index to remove final comma

            for cell in self.cells:
                f.write(f'{cell.id},{cell.frequency},{cell.percent_loc},')
                f.write(f'{cell.loc[1]},{cell.loc[2]},{cell.loc[3]},{cell.volume},{cell.summed},')

                for id in cell.channel_names:
                    f.write(
                        f'{cell.channel_stats[id]["mean"]},{cell.channel_stats[id]["median"]},{cell.channel_stats[id]["std"]},{cell.channel_stats[id]["var"]},')
                    f.write(
                        f'{cell.channel_stats[id]["min"]},{cell.channel_stats[id]["max"]},{cell.channel_stats[id]["%zero"]},{cell.channel_stats[id]["%saturated"]},')
                f.write('\n')
            f.close()
        elif self.analysis_type == 'detect':
            label = 'cellID,type,score,frequency,percent_loc,x_loc,y_loc'

            if filename is None and self.path is not None:
                filename = os.path.splitext(self.path)[0] + '.csv'  # Remove .lif and add .csv
            elif filename is None and self.path is None:
                filename = 'analysis.csv'
            f = open(filename, 'w')
            f.write(f'Filename: {self.filename}\n')
            f.write(f'Analysis Date: {self.analysis_date}\n')
            f.write(label[:-1:] + '\n')  # index to remove final comma

            for cell in self.cells:
                f.write(f'{cell.id},{cell.type},{cell.scores},{cell.frequency},{cell.percent_loc},')
                f.write(f'{cell.loc[1]},{cell.loc[2]}')
                f.write('\n')
            f.close()

    # @graceful_exit('\x1b[1;31;40m' + 'ERROR: Figure Render failed.' + '\x1b[0m')
    def make_segment_fig(self, filename: Optional[str] = None) -> None:
        """
        Make summary figure for quick interpretation of results.
        :param filename: filename to save figure as. If unset, uses image filename.
        :return: None
        """
        fig = plt.figure(figsize=(15, 8))

        ax = []
        x = [0.05, 0.35]
        y = [0.55, 0.075]
        w = 0.2
        h = 0.375
        for i in range(2):
            ax_ = []
            for j in range(2):
                ax_.append(fig.add_axes([x[i], y[j], w, h]))
            ax.append(ax_)

        hist_ax = []
        x = [0.05 + 0.2, 0.35 + 0.2]
        y = [0.55, 0.075]
        w = 0.05
        h = 0.375
        for i in range(2):
            ax_ = []
            for j in range(2):
                ax_.append(fig.add_axes([x[i], y[j], w, h]))
            hist_ax.append(ax_)

        im_ax = fig.add_axes([0.65, 0.075 * 2 + 0.25, 0.33, 0.55 - 0.075 / 2])
        cgram_ax = fig.add_axes([0.652, 0.075, 0.33, 0.25])

        num_cell = []
        # fig.set_size_inches(11,8)
        ind = [[0, 0], [0, 1], [1, 0], [1, 1]]

        for i, key in enumerate(self.cells[0].channel_names):
            signal = []
            perc = []
            x, y = ind[i]
            if i > 3:
                continue

            print(len(self.cells))
            for cell in self.cells:
                if cell.percent_loc is not None:
                    perc.append(cell.percent_loc)
                signal.append(cell.channel_stats[key]['mean'])

            if len(perc) != len(signal):
                perc = [i for i in range(len(signal))]

            perc = torch.tensor(perc) / max(perc)
            signal = torch.tensor(signal)
            channel = key
            signal = signal.div(signal.max() * 2)

            mean_gfp = []
            try:
                for window in torch.linspace(0, 1, 20):
                    mean_gfp.append(signal[torch.logical_and(perc > window, perc < (window + 0.1))].mean())
            except IndexError:
                print(window, len(mean_gfp), signal.shape, perc.shape)

            stylestr = ['green', 'grey', 'blue', 'red']

            ax[x][y].plot(perc * 100, signal, 'ko', alpha=0.5)
            print(torch.linspace(0, 100, 20))
            ax[x][y].plot(torch.linspace(0, 100, 20), mean_gfp, c=stylestr[i], ls='-', lw=3)
            ax[x][y].set_xlabel('Cell Location (Cochlear % Apex -> Base)')
            ax[x][y].set_ylabel(f'Cell Mean {channel[i]} Intensity (AU)')
            ax[x][y].legend(['Cell'])
            ax[x][y].axhline(torch.mean(signal), c=stylestr[i], lw=1, ls='-')
            ax[x][y].axhline(torch.mean(signal) + torch.std(signal), c=stylestr[i], lw=1, ls='--', alpha=0.5)
            ax[x][y].axhline(torch.mean(signal) - torch.std(signal), c=stylestr[i], lw=1, ls='--', alpha=0.5)
            # ax[x][y].set_ylim([torch.mean(signal) - torch.std(signal)*3, torch.mean(signal) + torch.std(signal) * 3])

            hist_ax[x][y].hist(signal.numpy(), color=stylestr[i], bins=30, orientation='horizontal')
            hist_ax[x][y].spines['right'].set_visible(False)
            hist_ax[x][y].spines['top'].set_visible(False)
            # hist_ax[x][y].spines['bottom'].set_visible(False)
            hist_ax[x][y].spines['left'].set_visible(False)
            hist_ax[x][y].set_yticks([], minor=True)
            hist_ax[x][y].axis('off')

        # new = fig.add_axes([0, 0, 0.5, 0.5])
        # new.plot([0,1,2,3,4], [2,3,4,5,6])
        was_compressed = False
        if isinstance(self.mask, bytes):
            self.decompress_mask()
            was_compressed = True

        mask = self.mask.sum(-1).gt(0).squeeze(0)

        if was_compressed:
            self.compress_mask()

        im_ax.imshow(mask.numpy())

        if self.curvature is not None:
            im_ax.plot(self.curvature[0, :], self.curvature[1, :], lw=3)
            apex = self.curvature[:, self.cochlear_distance.argmax()]
            im_ax.plot(apex[0], apex[1], 'ro')
        im_ax.set_title('Max Projection Cell Masks')

        perc = []
        for cell in self.cells:
            if cell.percent_loc is not None:
                perc.append(cell.percent_loc)
        perc = torch.tensor(perc) / max(perc)

        fig.suptitle(os.path.split(self.filename)[-1])
        perc = perc.mul(100).float().round().div(100)
        cgram_ax.hist(perc.numpy(), bins=50, color='k')
        cgram_ax.set_title('Cochleogram')
        cgram_ax.set_xlabel('Cell Location (Cochlear % Apex -> Base)')
        cgram_ax.set_ylabel('Num Cells')

        if filename is not None:
            plt.savefig(filename)
        else:
            plt.show()

        return fig

    def make_detect_fig(self, image: Tensor, filename: Optional[str] = None):
        """
        image has to be torch.Tensor

        :param image:
        :param filename:
        :return:
        """
        print('Rendering Figure...', end='')
        image = image.cpu() if image.min() < 0 else image.cpu()

        while image.shape[0] < 3:
            zeromat = torch.zeros((1, image.shape[1], image.shape[2]))
            image = torch.cat((zeromat, image), dim=0)

        # image = torchvision.transforms.functional.adjust_brightness(image, factor)
        # image = torchvision.transforms.functional.adjust_contrast(image, factor*0.65)

        _, x, y = image.shape
        ratio = x / y
        fig = plt.figure(figsize=(y / 200, x / 200))
        ax = plt.Axes(fig, [0., 0., 1., 1.])
        ax.set_axis_off()
        fig.add_axes(ax)

        blue = '#2b75ff'

        ax.imshow(image.numpy().transpose((1, 2, 0)))
        #
        # Loop over all cells and plot them on an image
        if self.curvature is not None:
            ax.plot(self.curvature[0, :].numpy(), self.curvature[1, :].numpy(), '-', color=blue)
            for i, cell in enumerate(self.cells):
                _, x0, y0, _ = cell.loc
                cell.calculate_frequency(self.curvature, self.cochlear_distance)
                x1, y1 = self.curvature[:, cell._curve_ind]

                marker = 's' if cell.type == 'IHC' else 'o'
                # color = '#ff2b2b' if marker == 's' else '#36ff3c'  # 1B512D
                color = '#FFFF00' if marker == 's' else '#66ff00'  # 1B512D
                ax.plot([x0, x1], [y0, y1], '-', color=blue, linewidth=0.3, alpha=0.66)
                ax.plot([x0], [y0], marker, color=color, markersize=1)
                f1 = cell.frequency

                # Put diagnostic text on the image
                s = f'ID: {cell.id}\nFreq: {f1:.2f} kHz\n{self.cochlear_distance[cell._curve_ind]:.2f} μm'
                ax.text(x0, y0, s, fontsize=1.5, c='w')
        else:
            for i, cell in enumerate(self.cells):
                _, x0, y0, _ = cell.loc
                marker = 's' if cell.type == 'IHC' else 'o'
                # color = '#ff2b2b' if marker == 's' else '#36ff3c'  # 1B512D
                color = '#FFFF00' if marker == 's' else '#66ff00'  # 1B512D
                ax.plot([x0], [y0], marker, color=color, markersize=1)

                # Put diagnostic text on the image
                s = f'ID: {cell.id}'
                ax.text(x0, y0, s, fontsize=1.5, c='w')

            # # What is the cell freq from ground truth
        if filename is not None:
            fig.savefig(filename, dpi=400)
        else:
            fig.savefig(self.path[:-4:] + '.jpg', dpi=400)
        plt.close(fig)
        print('DONE')

    def make_cochleogram(self, filename: Optional[str] = None, type: Optional[str] = None):

        if self.curvature is None:
            print('\x1b[1;33;40mWARNING: ' +
                  'Predicted Cochlear Distance is below 4000um. Not sufficient information to generate cochleogram.'
                  + '\x1b[0m')
            return None

        def hist_coords(dist, nbin=50):
            dist = torch.tensor(dist)
            hist = torch.histc(dist.float(), nbin)

            b = torch.linspace(dist.min(), dist.max(), nbin)
            b = torch.linspace(dist.min(), b[-2].item(), nbin)

            bin_x = [dist.min(), b[0].item()]
            bin_y = [0, hist[0].item()]
            for i in range(1, len(b)):
                bin_x.append(b[i].item())
                bin_x.append(b[i].item())
                bin_y.append(hist[i - 1].item())
                bin_y.append(hist[i].item())
            bin_x.append(dist.max())
            bin_x.append(dist.max())
            bin_y.append(bin_y[-1])
            bin_y.append(0)
            return bin_x, bin_y

        nbin = 100
        plt.figure(figsize=(4, 2))
        plt.ylim([0, 50])
        # GT curve
        ###############
        dist = []
        for cell in self.cells:
            cell.calculate_frequency(self.curvature, self.cochlear_distance)
            if type is not None and cell.type == type:
                dist.append(cell.distance)
            elif type is None:
                dist.append(cell.distance)

        x, y = hist_coords(dist, nbin)
        plt.plot(x, y, '-')
        plt.xlabel(r'Cochlear Distance $Base\ →\ Apex\ (\mu m)$')
        plt.ylabel('Hair Cells')
        # plt.legend(['Ground Truth', 'Predicted'], frameon=False)
        ax = plt.gca()
        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)
        ax.yaxis.set_ticks_position('left')
        ax.xaxis.set_ticks_position('bottom')
        plt.tight_layout()
        if filename is not None:
            plt.savefig(filename, dpi=450)
        else:
            plt.savefig(self.path[:-4:] + '_cochleogram.pdf', dpi=450)

        fig = plt.gcf()
        plt.close(fig)

        return fig


if __name__ == '__main__':
    a = torch.randint(0, 255, (1, 1000, 1000, 45))
    a[:, 2500::, :, :] = 0
    print('creating object')
    c = Cochlea(mask=a, cells=torch.arange(0, 255))
    c.render_mask('test.tiff')



import torch
import torch.nn as nn
from src.modules import RDCBlock, Up, Down, f
from warnings import filterwarnings

filterwarnings("ignore", category=UserWarning)

class RDCNet(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(RDCNet, self).__init__()

        complexity = 10

        self.strided_conv = nn.Conv3d(in_channels, complexity, kernel_size=3, stride=2, padding=1)
        self.RDCblock = RDCBlock(complexity)
        self.out_conv = nn.Conv3d(complexity, out_channels=complexity, kernel_size=3,padding=1)
        self.transposed_conv = nn.ConvTranspose3d(in_channels=complexity, out_channels=out_channels,
                                                  stride=(2,2,2), kernel_size=(4, 4, 4), padding=(1,1,1))

    def forward(self, x):
        x = self.strided_conv(x)
        for t in range(10):
            if t == 0:
                y = torch.zeros(x.shape).cuda()
            in_ = torch.cat((x, y), dim=1)
            y = self.RDCblock(in_) + y
        y = self.out_conv(y)
        return self.transposed_conv(y)


class RecurrentUnet_fail(nn.Module):

    def __init__(self, in_channels=4, out_channels=3):
        super(RecurrentUnet_fail, self).__init__()
        channels = [8,16,32,64,128, 256]
        channels = [5,6,7,8,16,32]

        self.conv9 = nn.Conv3d(in_channels, channels[0], kernel_size=9, padding=4)
        self.conv7 = nn.Conv3d(channels[0], channels[1], kernel_size=7, padding=3)
        self.conv5 = nn.Conv3d(channels[1], channels[2], kernel_size=5, padding=2)
        self.conv3 = nn.Conv3d(channels[2], channels[3], kernel_size=3, padding=1)

        self.intermediate_conv = nn.Conv3d(channels[3]*2, channels[3], kernel_size=1,padding=0)

        kernel = {'conv1': (3, 3, 3), 'conv2': (3, 3, 3)}
        upsample_kernel = (6, 6, 5)
        max_pool_kernel = (2, 2, 1)
        upsample_stride = (2, 2, 1)
        dilation = {'conv1': 1, 'conv2': 1}
        groups = {'conv1': 1, 'conv2': 1}

        # f_z
        self.down2_fz = Down(in_channels=channels[3], out_channels=channels[4],
                             kernel=kernel, dilation=dilation, groups=groups,padding=1)
        self.down3_fz = Down(in_channels=channels[4], out_channels=channels[5],
                             kernel=kernel, dilation=dilation, groups=groups,padding=1)
        self.up1_fz = Up(in_channels=channels[5], out_channels=channels[4], kernel=kernel, dilation=dilation, groups=groups,
                         upsample_kernel=upsample_kernel, upsample_stride = upsample_stride, padding_down=1,padding_up=2)

        # f_h
        self.down2_fh = Down(in_channels=channels[3], out_channels=channels[4],
                             kernel=kernel, dilation=dilation, groups=groups, padding=1)
        self.down3_fh = Down(in_channels=channels[4], out_channels=channels[5],
                             kernel=kernel, dilation=dilation, groups=groups, padding=1)
        self.up1_fh = Up(in_channels=channels[5], out_channels=channels[4], kernel=kernel, dilation=dilation, groups=groups,
                         upsample_kernel=upsample_kernel, upsample_stride=upsample_stride, padding_down=1, padding_up=2)

        self.out_conv = nn.Conv3d(channels[4], out_channels, 1)
        self.tanh = nn.Tanh()
        self.sigmoid = nn.Sigmoid()
        self.max_pool = nn.MaxPool3d(max_pool_kernel)

        self.fz = f(self.down2_fz, self.down3_fz, self.up1_fz, self.max_pool)
        self.fh = f(self.down2_fh, self.down3_fh, self.up1_fh, self.max_pool)

    def forward(self, image):
        image = self.conv9(image)
        image = self.conv7(image)
        image = self.conv5(image)
        image = self.conv3(image)
        image = self.max_pool(image)

        for t in range(10):
            if t == 0:
                s_t = torch.zeros(image.shape).cuda()

            x = torch.cat((image, s_t), dim=1)
            x = self.intermediate_conv(x)

            # Recurrent Bit!!!
            h = self.tanh(self.fh(x))
            if t == 0:
                h_t = torch.ones(h.shape).cuda()

            z = self.sigmoid(self.fz(x))
            h_t = (h_t * z) + (-1 * z * h)

            s_t = x

        return self.out_conv(x)


import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    description="A Hair Cell Analysis Toolbox",
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    python_requires='>=3',
    entry_points={'console_scripts': ['hcat-segment=hcat.main:segment',
                                      'hcat-detect=hcat.main:detect',
                                      'hcat=hcat.main:cli']},
    install_requires=[
        'auto_mix_prep>=0.2.0',
        'click>=8.1.3',
        'elasticdeform>=0.4.9',
        'GPy>=1.10.0',
        'kornia>=0.6.5',
        'lz4>=4.0.1',
        'matplotlib>=3.5.1',
        'numpy>=1.22.3',
        'pandas>=1.4.2',
        'Pillow>=9.2.0',
        'psutil>=5.9.1',
        'PySimpleGUI>=4.60.1',
        'scikit_image>=0.19.2',
        'scikit_learn>=1.1.1',
        'scipy>=1.7.3',
        'setuptools>=61.2.0',
        'timm>=0.5.4',
        'torch>=1.12.0',
        'torchvision>=0.13.0',
        'tensorboard>=2.9.1',
        'tqdm>=4.64.0',
        'wget>=3.2',
    ]
)



from hcat.detect import _detect
from hcat.segment import _segment
from hcat.detect_gui import gui

import glob

import click
from typing import Optional
import warnings
warnings.filterwarnings("ignore")

"""
This is the main entry point to all analysis scripts. Each analysis script is found in hcat.
"""


@click.group(invoke_without_command=True)
@click.pass_context
def cli(ctx):
    if ctx.invoked_subcommand is None:
        """ Default subcommand is running full analysis """
        gui().main_loop()


@cli.command()
@click.argument('f', default=None)
@click.option('--channel', default=2, help='Channel index to segment')
@click.option('--intensity_reject_threshold', default=0.05, help='Cell cytosol intensity rejection threshold')
@click.option('--dtype', default=None, help='dtype of input image')
@click.option('--unet', is_flag=True, help='Run with Unet+Watershed Backbone')
@click.option('--cellpose', is_flag=True, help='Run with Cellpose Backbone')
@click.option('--figure', is_flag=True, help='Save preliminary analysis figure')
@click.option('--no_post', is_flag=True, help='Do not postprocess')
def segment(f: str, channel: int, intensity_reject_threshold: float,
            dtype: Optional[str],
            unet: bool, cellpose: bool, no_post: bool,
            figure: Optional[bool]) -> None:
    raise NotImplementedError('Segmentation currently in not supported.')
    # files = glob.glob(f)
    # for filename in files:
    #     try:
    #         _segment(f=filename, channel=channel, intensity_reject_threshold=intensity_reject_threshold,
    #                  dtype=dtype, unet=unet, cellpose=cellpose, no_post=no_post, figure=figure)
    #     except Exception:
    #         print(f'Error in file: {filename}')

    # return None


@cli.command()
@click.argument('f', default=None)
@click.option('--curve_path', default=None, help='CSV path of manually annotated cochlear curvature')
@click.option('--cell_detection_threshold', default=0.85, help='Threshold (between 0 and 1) of cell detection.')
@click.option('--nms_threshold', default=0.1, help='Allowable overlap between prediction boxes.')
@click.option('--dtype', default=None, help='dtype of input image')
@click.option('--save_xml', is_flag=True, help='Threshold (between 0 and 1) of cell detection.')
@click.option('--save_png', is_flag=True, help='Save a png for annoation purposes.')
@click.option('--save_fig', is_flag=True, help='Threshold (between 0 and 1) of cell detection.')
@click.option('--normalize', is_flag=True, help='Threshold (between 0 and 1) of cell detection.')
@click.option('--pixel_size', default=None, help='Pixel size in nm')
@click.option('--cell_diameter', default=None, help='Cell diameter in pixels')
def detect(f: str, curve_path, cell_detection_threshold, nms_threshold, save_xml, save_png,
           save_fig, normalize, pixel_size, dtype, cell_diameter):

    cell_diameter = float(cell_diameter) if cell_diameter is not None else None
    # Evaluate a single image

    files = glob.glob(f)
    for filename in files:
        try:
            _detect(f=filename,
                    curve_path=curve_path,
                    cell_detection_threshold=cell_detection_threshold,
                    nms_threshold=nms_threshold,
                    save_png=save_png,
                    save_xml=save_xml,
                    save_fig=save_fig,
                    normalize=normalize,
                    pixel_size=pixel_size,
                    cell_diameter=cell_diameter,
                    dtype=dtype)
        except Exception as e:
            print(f'Critical Error! Aborting - {filename}')
            raise e


if __name__ == '__main__':
    cli()



import torch.nn as nn
import torch
from src.train.transforms import _crop
from typing import List


class r_unet_LTS(nn.Module):
    """
    LTS r_unet. Assumes 512 scale param

    """
    def __init__(self, in_channels=1, out_channels=4):
        super(r_unet_LTS, self).__init__()

        self.activation = nn.LeakyReLU()

        # c = [20, 25, 30]
        c = [10, 20, 30]

        self.down_1 = r_block(in_channels, c[0], 5, 1, 1)
        self.down_2 = r_block(c[0], c[1], 5, 2, 2)
        self.down_3 = r_block(c[1], c[2], 5, 3, 3)

        self.up_1 = r_block(c[2] + c[1], c[1], 5)
        self.up_2 = r_block(c[1] + c[0], c[0] + 10, 5)

        self.out_embed = nn.Conv3d(c[0] + 10, 3, kernel_size=3, stride=1, dilation=1, padding=1)
        self.out_prob = nn.Conv3d(c[0] + 10, 1, kernel_size=3, stride=1, dilation=1, padding=1)

        self.stride_1 = nn.Conv3d(c[0], c[0], kernel_size=3, stride=(3, 3, 2), dilation=1, padding=0)
        self.stride_2 = nn.Conv3d(c[1], c[1], kernel_size=3, stride=(3, 3, 2), dilation=1, padding=0)

        self.transpose_1 = nn.ConvTranspose3d(c[2], c[2], kernel_size=3, stride=(3, 3, 2), padding=0, dilation=1)
        self.transpose_2 = nn.ConvTranspose3d(c[1], c[1], kernel_size=3, stride=(3, 3, 2), padding=0, dilation=1)

        self.tanh = nn.Tanh()
        self.sigmoid = nn.Sigmoid()
        self.softmax = nn.Softmax()

    def forward(self, x: torch.Tensor) -> torch.Tensor:

        x = self.down_1(x)
        y = self.down_2(self.activation(self.stride_1(x)))
        z = self.down_3(self.activation(self.stride_2(y)))
        z = self.activation(self.transpose_1(z))
        z = self.up_1(torch.cat((crop(y, z.shape), z), dim=1))
        z = self.activation(self.transpose_2(z))
        z = self.up_2(torch.cat((crop(x, z.shape), z), dim=1))

        z = torch.cat((
            self.tanh(self.out_embed(z)),
            self.sigmoid(self.out_prob(z))
        ), dim=1)

        return z


class r_block(nn.Module):
    """

    Recurrently applies a block to an input and outputs later

    """
    def __init__(self, in_channels: int, out_channels: int, n_loop: int, dilation: int = 1, padding: int = 1):
        super(r_block, self).__init__()

        self.activation = nn.LeakyReLU()
        self.n = n_loop

        self.conv1x1 = nn.Conv3d(in_channels * 2, in_channels, kernel_size=1)
        # self.conv7x7_3 = nn.Conv3d(in_channels, in_channels, kernel_size=(3,3,1), padding=(1,1,0), stride=1, dilation=1)
        self.conv7x7_1 = nn.Conv3d(in_channels, in_channels, kernel_size=3, padding=padding, stride=1, dilation=dilation)
        self.conv7x7_2 = nn.Conv3d(in_channels, out_channels, kernel_size=3, padding=padding, stride=1, dilation=dilation)

        self.activation = nn.LeakyReLU()

        self.batch_norm_conv_1x1 = nn.BatchNorm3d(in_channels)
        self.batch_norm_conv_7x7_1 = nn.BatchNorm3d(in_channels)
        # self.batch_norm_conv_7x7_3 = nn.BatchNorm3d(in_channels)
        self.batch_norm_conv_7x7_2 = nn.BatchNorm3d(out_channels)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        y = torch.zeros(x.shape, device=x.device)
        for _ in range(self.n):
            y = torch.cat((x, y), dim=1)
            y = self.activation(self.batch_norm_conv_1x1(self.conv1x1(y)))
            # y = self.activation(self.batch_norm_conv_7x7_3(self.conv7x7_3(y)))
            y = self.activation(self.batch_norm_conv_7x7_1(self.conv7x7_1(y)))

        y = self.activation(self.batch_norm_conv_7x7_2(self.conv7x7_2(y)))

        return y


@torch.jit.script
def crop(x: torch.Tensor, shape: List[int]) -> torch.Tensor:
    return _crop(x, 0, 0, 0, int(shape[2]), int(shape[3]), int(shape[4]))
